import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import cv2
from PIL import Image
import numpy as np
from sklearn.metrics import confusion_matrix, accuracy_score
from seaborn import heatmap
from mpl_toolkits.mplot3d import Axes3D
# import os, fnmatch
import matplotlib.pyplot as plt
from matplotlib.patches import Ellipse

def imshow(image):
    pil_image = Image.fromarray(image)
    pil_image.show()

def get_image(path, show=False):
    img = cv2.imread(path)
    if show:
        imshow(img)
    return img

def accuracy(Ys,Y,st):
    print(st)
    if Y.shape[1]>Y.shape[0]:
        Y = Y.transpose()
        Ys = Ys.transpose()
    if Y.shape[1]>1:
        d   = np.argmax(Y,axis = 1)
        ds  = np.argmax(Ys,axis = 1)
    else:
        d  = Y
        ds = Ys
    C   = confusion_matrix(d,ds) 
    acc = accuracy_score(d,ds)
    print('Confusion Matrix:')
    print(C)
    print('Accuracy = '+str(acc))
    print()
    nm = C.shape[0]
    plt.figure(figsize=(7,5))
    heatmap(C, annot=True, fmt="d", cmap="YlGnBu")
    plt.xlim(0,nm)
    plt.ylim(nm,0)
    plt.title('Confusion Matrix '+st,fontsize=14)
    plt.show()
    return acc

def get_ellipse_path(params):
    cx    = params[0]
    cy    = params[1]
    a     = params[2]
    b     = params[3]
    alpha = params[4]
    ell   = Ellipse((cx,cy),a*2.,b*2.,alpha)
    coord = ell.get_verts()
    xs    = coord[:,0]
    ys    = coord[:,1]
    return xs,ys


def plot_ellipses(ellipse_list):
    '''Plot a list of ellipses'''
    fig=plt.figure()
    ax2=fig.add_subplot(111)
    n = len(ellipse_list)
    for i in range(n):
        cii = ellipse_list[i]
        x=cii[:,0]
        y=cii[:,1]
        if i==0:
            ax2.plot(x,y,'.')
        else:
            ax2.plot(x,y,'-')
    plt.show()

def plot_ellipses_image(I,x,y=0):
    if y==0:
        x,y  = get_ellipse_path(x)
    fig=plt.figure()
    ax2=fig.add_subplot(111)
    implot = plt.imshow(I,cmap='gray')    
    ax2.plot(x,y,'-',color='Red')
    plt.show()

def plot_confusion(confusion_mtx,acc,st,dmin,dmax,p100=True):
    C = confusion_mtx.copy()
    print('Confusion matrix - '+st+':')
    print(C)
    accst = f'Acc = {acc:.4f}'    
    print(accst)
    plt.figure(figsize=(10,8))
    n = C.shape[0]
    if p100:
        for i in range(n):
            si = np.sum(C[i,:])
            C[i,:] = C[i,:]/si*100
        heatmap(C, annot=True, fmt="d",cmap="YlGnBu",vmin=0, vmax=100)
        st = 'Confusion Matrix [%] - '+st+': '+accst
    else:
        heatmap(C, annot=True, fmt="d",cmap="YlGnBu")
        st = 'Confusion Matrix - '+st+': '+accst
    plt.xlim(dmin, dmax)
    plt.ylim(dmax, dmin)
    plt.title(st,fontsize=14)
    plt.show()



def show_confusion_matrix(Yt,Ys,st,categorical=0):
    # Yt : ground truth
    # Ys : prediction
    if categorical:
        if Yt.shape[1]>Yt.shape[0]:
            Yt = Yt.transpose()
            Ys = Ys.transpose()
        dt  = np.argmax(Yt,axis = 1)
        ds  = np.argmax(Ys,axis = 1)
    else:
        dt = Yt
        ds = Ys
    confusion_mtx = confusion_matrix(dt,ds) 
    acc = accuracy_score(dt,ds)
    dmin = np.min(dt)
    if np.min(dt)==1:
        dmin = dmin-1
        dmax = np.max(dt)
    else:
        dmax = np.max(dt)+1
        dmin = 0        
    plot_confusion(confusion_mtx,acc,st,dmin,dmax)    






def plot_confusion_matrix(dt,ds,st):
    print('computing confusion matrix...')
    print('WARNING: use show_confusion_matrix!')
    confusion_mtx = confusion_matrix(dt,ds) 
    print(confusion_mtx)
    plt.figure(figsize=(10,8))
    n = confusion_mtx.shape[0]
    for i in range(n):
        si = np.sum(confusion_mtx[i,:])
        confusion_mtx[i,:] = confusion_mtx[i,:]/si*100
    heatmap(confusion_mtx, annot=True, fmt="d",cmap="YlGnBu",vmin=0, vmax=100)
    # dmin = np.min(dt)
    if np.min(dt)==1:
        dmax = np.max(dt)
    else:
        dmax = np.max(dt)+1
        dmin = 0        
    plt.xlim(dmin, dmax)
    plt.ylim(dmax, dmin)
    acc = accuracy_score(dt,ds)    
    accst = f'Acc = {acc:.4f}' 
    print('Accuracy = '+accst)   
    plt.title('Confusion Matrix [%] - '+st+': '+accst,fontsize=14)
    plt.show()

def print_confusion(dt,ds):
    # dt - true, ds - predicted
    C   = confusion_matrix(dt,ds) 
    print('Confusion Matrix:')
    print(C)
    acc = accuracy_score(dt,ds) 
    print('Accuracy = '+str(acc))

def plot_decision_lines(clf,X,show=0,decisionline=1):
    # based on example of https://scikit-learn.org
    h = 0.05
    x_min, x_max = X[:, 0].min() - 1, X[:, 0].max() + 1
    y_min, y_max = X[:, 1].min() - 1, X[:, 1].max() + 1
    xx, yy = np.meshgrid(np.arange(x_min, x_max, h),
                         np.arange(y_min, y_max, h))
    if decisionline == 1:
        Z = clf.predict(np.c_[xx.ravel(), yy.ravel()])
        Z = Z.reshape(xx.shape)
        plt.contourf(xx, yy, Z, cmap=plt.cm.Reds, alpha=0.8)
    plt.xlim(x_min, x_max)
    plt.ylim(y_min, y_max)
    if show==1:
        plt.show()

def plot_features(X,d,st,show=1):
    dmin = int(np.min(d)) 
    dmax = int(np.max(d)) 
    for j in range(dmin,dmax+1):
        plt.scatter(X[d==j,0],X[d==j,1],label='Class '+str(j),cmap=plt.cm.autumn,s=17)
    plt.grid(True)
    plt.legend()
    plt.xlabel('$x_1$',fontsize=14)
    plt.ylabel('$x_2$',fontsize=14)
    plt.title('Feature Space - '+st,fontsize=14)
    if show==1:
        plt.show()

def plot_features2(X,d,st,show=1):
    # based on example of https://scikit-learn.org
    x_min, x_max = X[:, 0].min() - 1, X[:, 0].max() + 1
    y_min, y_max = X[:, 1].min() - 1, X[:, 1].max() + 1
    print(X.shape)
    plt.scatter(X[:, 0], X[:, 1], c=d, cmap=plt.cm.autumn)
    plt.xlabel('$x_1$',fontsize=14)
    plt.ylabel('$x_2$',fontsize=14)
    plt.xlim(x_min, x_max)
    plt.ylim(y_min, y_max)
    plt.xticks(())
    plt.yticks(())
    plt.title('Feature Space - '+st,fontsize=14)
    #plt.grid(True)
    #plt.legend()
    if show == 1:
        plt.show()

def plot_features3(X,d,st,show=1,view=(30,60)):
    fig = plt.figure()
    ax = fig.add_subplot(111, projection='3d')
    ax.scatter(X[:, 0], X[:, 1], X[:, 2], c=d, cmap=plt.cm.autumn)
    ax.set_xlabel('$x_1$',fontsize=14)
    ax.set_ylabel('$x_2$',fontsize=14)
    ax.set_zlabel('$x_3$',fontsize=14)
    plt.title('Feature Space - '+st,fontsize=14)
    ax.view_init(view[0],view[1])
    # plt.axis('off')
    # plt.grid(b=None)
    if show == 1:
        plt.show()

def plot_features_y(X,Y,st):

    if X.shape[1]<X.shape[0]:
        X = X.transpose()
    if Y.shape[1]<Y.shape[0]:
        Y = Y.transpose()
    if Y.shape[0]>1:
        d  = np.argmax(Y,axis = 0)
    else:
        d = Y
    K = np.max(d)+1
    color = ['blue', 'orange', 'green', 'red']
    for j in range(K):
        # plt.scatter(X[d==j,0],X[d==j,1],label='Class '+str(j),color='tab:'+color[j],s=1)
        plt.scatter(X[0,d==j],X[1,d==j],label='Class '+str(j),color='tab:'+color[j],s=17)

    plt.grid(True)
    plt.legend()
    plt.xlabel('$x_1$',fontsize=14)
    plt.ylabel('$x_2$',fontsize=14)
    plt.xticks(())
    plt.yticks(())
    plt.title('Feature Space - '+st,fontsize=14)
    plt.show()



def show_clf_results(clf,X,d,Xt,dt,d0,ds,st,decisionline=1):
    gs = gridspec.GridSpec(1, 2)
    fig = plt.figure(figsize=(18, 6))
    print('Training:')
    acc = accuracy_score(d,d0) 
    accst = f'Acc = {acc:.4f}'    
    ax = plt.subplot(gs[0,0])
    print_confusion(d,d0)                             # confusion matrix in training
    plot_decision_lines(clf,X,0,decisionline)          # decision lines
    plot_features(X,d,st+' - Training: '+accst,0)     # feature space in training   
    ax = plt.subplot(gs[0,1])
    print('Testing:')
    acc = accuracy_score(ds,dt) 
    accst = f'Acc = {acc:.4f}'    
    print_confusion(dt,ds)                            # confusion matrix in testing
    plot_decision_lines(clf,X,0,decisionline)          # decision lines
    plot_features(Xt,dt,st+' - Testing: '+accst,1)    # feature space in testing
    
def plot_ROC(fpr,tpr,label='ROC',auc=0,seq=[0, 1]):
    k = seq[0]
    n = seq[1]

    if auc>0:
        label = label + f': AUC = {auc:.4f}' 

    plt.plot(fpr, tpr,  label=label)
    plt.plot([0, 1], [0, 1], color='grey', linestyle='--')
    plt.xlabel('fpr')
    plt.ylabel('tpr')
    plt.title('ROC Curve')
    plt.grid(True)
    plt.legend()
    if k==n-1:
        plt.show()

def plot_precision_recall(precision,recall,label='precision/recall',ap=0,seq=[0, 1]):
    k = seq[0]
    n = seq[1]

    if ap>0:
        label = label + f': AP = {ap:.4f}' 

    plt.plot(recall, precision, label=label)
    
    plt.xlabel('recall')
    plt.ylabel('precision')
    plt.title('Precision/Recall Curve')
    plt.grid(True)
    plt.legend()
    if k==n-1:
        plt.show()



def plot_cnn_history(history):
    # loss curves
    print('displaying loss and accuracy curves...')
    plt.figure(figsize=[8,6])
    plt.plot(history.history['loss'],'r',linewidth=3.0)
    plt.plot(history.history['val_loss'],'b',linewidth=3.0)
    plt.legend(['Training loss', 'Validation Loss'],fontsize=18)
    plt.xlabel('Epochs ',fontsize=16)
    plt.ylabel('Loss',fontsize=16)
    plt.title('Loss Curves',fontsize=16)
    plt.show()

    # accuracy curves
    plt.figure(figsize=[8,6])
    plt.plot(history.history['acc'],'r',linewidth=3.0)
    plt.plot(history.history['val_acc'],'b',linewidth=3.0)
    plt.legend(['Training Accuracy', 'Validation Accuracy'],fontsize=18)
    plt.xlabel('Epochs ',fontsize=16)
    plt.ylabel('Accuracy',fontsize=16)
    plt.title('Accuracy Curves',fontsize=16)
    plt.show()

def plot_loss(loss_train):
    plt.figure(figsize=[8,6])
    plt.plot(loss_train,'r',linewidth=1.5)
    # plt.plot(loss_val,'b',linewidth=1.5)
    # plt.legend(['Training loss', 'Validation Loss'],fontsize=14)
    plt.xlabel('Epochs ',fontsize=14)
    plt.ylabel('Loss',fontsize=14)
    plt.title('Training Loss',fontsize=14)
    xmin, xmax = plt.xlim()
    ymin, ymax = plt.ylim()
    plt.xlim(0,xmax)
    plt.ylim(0,ymax)
    plt.grid(True)
    plt.show()



"""
This module contains all the methods related to images segmentation.

"""

__author__ = 'Christian Pieringer'
__created__ = '2019.12.26'

import numpy as np

from cv2 import MSER_create
from skimage.filters import threshold_otsu, gaussian
from skimage.measure import find_contours
from skimage.morphology import binary_closing
from skimage.morphology import binary_dilation
from skimage.morphology import disk
from skimage.morphology import flood
from skimage.morphology import remove_small_holes, remove_small_objects

from skimage.segmentation import clear_border, find_boundaries
from skimage.measure import label, regionprops

from pyxvis.processing.images import Edge 


def seg_bimodal(img, p=-0.1):
    """

    :param img: numpy array for a gray-scale input image.
    :param p: float to denote the offset of threshold with p between -1 and 1. A
                positive value is used to dilate the segmentation, the negative
                to erode.
    :return: a tuple with the binary image and the contour of the mask.

    """
    img_d = img.astype('double')
    imax = np.max(img_d.flatten())
    imin = np.min(img_d.flatten())
    img_j = (img_d - imin) / (imax - imin)  # Image normalization [0, 1]

    ni = int(np.fix(img_j.shape[0] / 4))
    nj = int(np.fix(img_j.shape[1] / 4))

    # An heuristic method to ensure that the segmented object will
    # be labeled as 1.
    if np.mean(img_j[0:ni, 0:nj].flatten()) > 0.3:
        img_j = 1 - img_j

    # Transform a double image to unit8
    img_jj = (255 * img_j).astype('uint8')

    # Apply Gaussian filtering to remove noise and compute Otsu thresholding.
    # This step improve the thresholding
    img_jj = gaussian(img_jj, 2.0)
    level = threshold_otsu(img_jj)
    mask = img_jj > (level + p)

    a = remove_small_objects(mask, np.fix(len(mask.flatten()) / 100))
    c = binary_closing(a.astype('double'), disk(7))

    # Why 8?
    r = remove_small_holes(c, 8)
    contours = find_contours(r, 0.1)

    return r, contours


def region_growing(img, seed_point, tolerance=20):
    """
    Args:
        img (ndarray): image
        seed_point (tuple): define the seed point.
        tolerance: a float or int value to define the maximal difference of grayvalues in the region. According to the
                   documentation in scikit-image, if tolerance is provided, adjacent points with values within plus 
                   or minus tolerance from the seed point are filled (inclusive).

    Returns:
        mask (ndarray): output binary image
    """
    mask = flood(img, seed_point, tolerance=tolerance)
    mask = binary_dilation(mask, np.ones((3, 3)))

    return mask


def seg_log_feature(_img, _mask=None, area=None, gray=None, contrast=None, sigma=None):
    """
    
    """
    if _mask.all():
        _mask = np.ones(_img.shape)

    # Unpackage parameters
    amin = area[0]
    amax = area[1]
    gmin = gray[0]
    gmax = gray[1]
    cmin = contrast[0]
    cmax = contrast[1]
    
    se = disk(3)  # Structural element
    re = binary_dilation(_mask, se)

    e = clear_border(find_boundaries(_mask, connectivity=1, mode='inner')).astype(np.uint8)
    edge = Edge('log', 1e-8, sigma)  # Edge detection using log
    edge.fit(_img)

    b = edge.edges.astype(np.uint8)
    b = np.bitwise_and(b, re)
    b = np.bitwise_or(b, e)
    b = clear_border(np.bitwise_not(b))

    f, m = label(b, connectivity=1, return_num=True)  # Connectivity=1 == 4 Neighboors

    d = np.zeros(_img.shape).astype(np.uint8)
    
    for i in range(m):
        r = f == i
        b = np.bitwise_xor(r, binary_dilation(r, np.ones((17, 17)))).astype(np.uint8)
        ir = r == 1
        ib = b == 1
        area = np.sum(ir.flatten())
        gray_r = np.mean(_img[ir].flatten())
        gray_b = np.mean(_img[ib].flatten())
        contrast = gray_r / gray_b

        if (area >= amin) and (area <= amax) and (gray_r >= gmin) and (gray_r <= gmax) and (contrast >= cmin) and (contrast <= cmax):
            d = np.bitwise_or(d, r)

    f, m = label(d, connectivity=1, return_num=True)  # Connectivity=1 == 4 Neighboors

    return f, m


def seg_mser(_img, area=(60, 20000), min_div=0.2, max_var=0.25, delta=5, area_threshold=200):
    """
    Image segmentation using MSER algorithm.

    Args:
        _img (ndarray): input image (uint8)
        area (int tuple): Range of areas to prune (min_area, max_area)
        min_div (float): Trace back to cut off MSER with diversity less than min_diversity
        max_var (float): Prune the area that have simliar size to its children
        delta (int): Controls how the stability is calculated comparing (sizei - sizei-delta) / sizei-delta. [1]
        area_threshold (float): the area threshold to cause re-initialize in color images [1]

    Returns:
       mask (ndarray): Binary image with detected objects
       label_regions (ndarray): Image with labels of the detected objects
       bboxes (ndarray): Bounding boxes for each detection in label_regions (x0, y0, width, height), where (x0, y0) is
                         the uppper-left corner.
    

    [1] Please check the OpenCV documentation for more details.
    """
    
    mser_opts = {
        '_min_area': area[0],
        '_max_area': area[1],
        '_max_variation': max_var,
        '_min_diversity': min_div,
        '_delta': delta,
        '_area_threshold': area_threshold
    }
    
    mser = MSER_create(**mser_opts)
    regions, bbox = mser.detectRegions(_img)
    
    mask = np.zeros(_img.shape)

    for dd, region in enumerate(regions):
        mask[region[:,1], region[:,0]] = 1

    label_regions, m = label(mask, connectivity=1, return_num=True)
    props = regionprops(label_regions)
    bboxes = np.array([[p.bbox[1], p.bbox[0], p.bbox[3] - p.bbox[1], p.bbox[2] - p.bbox[0]] for p in props])
    
    return mask, label_regions, bboxes



# gan_main.py
# GAN for images of NxN pixels (grayvalues)
# (c) D.Mery, 2019-2020
#
# Example for Training: For generating shuriken images of 32x32 pixels in 2500 epochs (and saving models and images each 250 epochs):
# python3 gan_main.py 0 shuriken_32x32.npy 2500 250
#
# For generating 1000 synthetic images of shuriken of 32x32 pixels using model learned after 1000 epochs:
# python3 gan_main.py 1 32 models/gan_model_002500.h5 1000
#
# See readme.txt for more examples


from keras.layers import Input, Dense, Reshape, Flatten, Dropout
from keras.layers import BatchNormalization, Activation, ZeroPadding2D
from keras.layers.advanced_activations import LeakyReLU
from keras.layers.convolutional import UpSampling2D, Conv2D
from keras.models import Sequential, Model
from keras.optimizers import Adam
import h5py
import matplotlib.pyplot as plt
import sys
import numpy as np
import os.path

# Constants
GAN_PATH = '../output/GAN/'

# based on implementation of
# https://github.com/eriklindernoren/Keras-GAN/blob/master/dcgan/dcgan.py

class DCGAN():
    def __init__(self,st):
        # load training data 
        if os.path.isfile(st):
            path_file = st
            self.X = load_train_defects(path_file)
            self.gansize = self.X.shape[1]
        else: 
            print(st)
            self.gansize = int(st)
            print(self.gansize)
        self.nmax = 3 # 3 layers in generator and discriminator

        # Input shape
        self.img_rows = self.gansize
        self.img_cols = self.gansize
        self.channels = 1
        self.img_shape = (self.img_rows, self.img_cols, self.channels)
        self.latent_dim = 100
        self.path_file = ''

        optimizer = Adam(0.0002, 0.5)

        # Build and compile the discriminator
        self.discriminator = self.build_discriminator()
        self.discriminator.compile(loss='binary_crossentropy',
            optimizer=optimizer,
            metrics=['accuracy'])

        # Build the generator
        self.generator = self.build_generator()

        # The generator takes noise as input and generates images
        z = Input(shape=(self.latent_dim,))
        img = self.generator(z)

        # For the combined model we will only train the generator
        self.discriminator.trainable = False

        # The discriminator takes generated images as input and determines validity
        valid = self.discriminator(img)

        # The combined model  (stacked generator and discriminator)
        # Trains the generator to fool the discriminator
        self.combined = Model(z, valid)
        self.combined.compile(loss='binary_crossentropy', optimizer=optimizer)

    def build_generator(self):

        model = Sequential()
        
        [p,d] = generator_size(self.gansize,self.nmax)
        
        n = len(d)

        model.add(Dense(d[0] * p * p, activation="relu", input_dim=self.latent_dim))
        model.add(Reshape((p, p, 128)))
        
        for i in range(n):
            model.add(UpSampling2D())
            model.add(Conv2D(d[i], kernel_size=3, padding="same"))
            model.add(BatchNormalization(momentum=0.8))
            model.add(Activation("relu"))
            
        model.add(Conv2D(self.channels, kernel_size=3, padding="same"))
        model.add(Activation("tanh"))

        model.summary()

        noise = Input(shape=(self.latent_dim,))
        img = model(noise)

        return Model(noise, img)

    def build_discriminator(self):

        model = Sequential()

        # model.add(ZeroPadding2D(padding=((0,1),(0,1)))) it was in after 64's layer

        d = 32

        model.add(Conv2D(d, kernel_size=3, strides=2, input_shape=self.img_shape, padding="same"))
        model.add(LeakyReLU(alpha=0.2))
        model.add(Dropout(0.25))

        for i in range(self.nmax):
            d = 2*d
            model.add(Conv2D(d, kernel_size=3, strides=2, padding="same"))
            model.add(BatchNormalization(momentum=0.8))
            model.add(LeakyReLU(alpha=0.2))
            model.add(Dropout(0.25))

        model.add(Flatten())
        model.add(Dense(1, activation='sigmoid'))

        model.summary()

        img = Input(shape=self.img_shape)
        validity = model(img)

        return Model(img, validity)

    def train(self, epochs, batch_size=128, save_interval=50):

        # Training data
        X_train = np.expand_dims(self.X, axis=3)

        # Adversarial ground truths
        valid = np.ones((batch_size, 1))
        fake = np.zeros((batch_size, 1))

        for epoch in range(epochs+1):

            # ---------------------
            #  Train Discriminator
            # ---------------------

            # Select a random half of images
            idx = np.random.randint(0, X_train.shape[0], batch_size)
            images = X_train[idx]

            # Sample noise and generate a batch of new images
            noise = np.random.normal(0, 1, (batch_size, self.latent_dim))
            gen_images = self.generator.predict(noise)

            # Train the discriminator (real classified as ones and generated as zeros)
            d_loss_real = self.discriminator.train_on_batch(images, valid)
            d_loss_fake = self.discriminator.train_on_batch(gen_images, fake)
            d_loss = 0.5 * np.add(d_loss_real, d_loss_fake)

            # ---------------------
            #  Train Generator
            # ---------------------

            # Train the generator (wants discriminator to mistake images as real)
            g_loss = self.combined.train_on_batch(noise, valid)


            # Plot the progress
            print ("%d [D loss: %f, acc.: %.2f%%] [G loss: %f]" % (epoch, d_loss[0], 100*d_loss[1], g_loss))

            # If at save interval => save generated image samples
            if epoch % save_interval == 0:
                self.save_gan_examples(epoch)

            if epoch % (save_interval) == 0:
                st = '%0*d' % (6,epoch)
                self.save_synthetic_images(st,1000)
                self.save_gan_model(st)

    def save_gan_examples(self, epoch=-1):
        r, c = 5, 5
        t = r*c
        if epoch >= 0: 
            st = '%0*d' % (6,epoch)
        else:
            st = 'output'
        st_path = GAN_PATH + 'examples/gan_examples_' + st + '.png'
        print('> saving '+str(t)+ ' synthetic images (examples) in ' + st_path)

        noise = np.random.normal(0, 1, (t, self.latent_dim))
        gen_images = self.generator.predict(noise)

        # Rescale images 0 - 1
        gen_images = 0.5 * gen_images + 0.5

        fig, axs = plt.subplots(r, c)
        cnt = 0
        for i in range(r):
            for j in range(c):
                axs[i,j].imshow(gen_images[cnt, :,:,0], cmap='gray')
                axs[i,j].axis('off')
                cnt += 1

        fig.savefig(st_path)
        plt.close()


    def save_synthetic_images(self, st, N):
        noise = np.random.normal(0, 1, (N, self.latent_dim))
        gen_images = self.generator.predict(noise)
        # print('> computing '+str(N) + ' GAN synthetic images...')

        # Rescale images 0 - 1
        gen_images = 0.5 * gen_images + 0.5
        save_images_gan(N,gen_images,st)


    def save_gan_model(self,st):
        st_path = GAN_PATH + 'models/gan_model_'+st+'.h5'
        print('> saving model in ' + st_path + ' ...')
        save_model(self.combined,st_path)

    def load_gan_model(self,st_path):
        # st_path = 'models/gan_model_'+st+'.h5'
        print('loading ' + st_path + ' ...')
        load_model(self.combined,st_path)

def save_images_gan(N,gen_images,st):
        st_path = GAN_PATH + 'synthetic/gan_synthetic_'+st
        print('> saving '+str(N)+' GAN synthetic images in '+st_path+'.npy')
        np.save(st_path,gen_images)
        return

def load_train_defects(st_file):
        print('loading training defect pacthes '+st_file+'...')
        x_train = np.load(st_file)
        return x_train

# from https://github.com/keras-team/keras/issues/10608

def save_model(model, filename):
    f = h5py.File(filename,'w')
    for symbolic_weights, weights in zip(model.weights, model.get_weights()):
        name = symbolic_weights.name
        f.create_dataset(name, data=weights)
    f.flush()
    f.close()
    
def load_model(model, filename):
    f = h5py.File(filename, 'r')
    weights = []
    for symbolic_weights in model.weights:
        name = symbolic_weights.name
        dataset = f[name]
        weights.append( dataset[:] )
    model.set_weights(weights)
    f.close()

def generator_size(p,nmax):

    n = 0
    ok = 0

    while not(ok):
        t = p/2
        if 2*np.fix(t)==p:
            n = n+1
            p = t
            if n == nmax:
                ok = 1
        else:
            ok = 1
        
    d = np.zeros((n,), dtype=int)
    g = 128
    for i in range(n):
        d[i] = g
        g = g/2

    p = int(p)
        # d = int(d)
        
    print(p)

    print(d)

    print(n)

    return p,d





import keras
import cv2
import os                   
import numpy                as np 
import matplotlib.pyplot    as plt
import tensorflow           as tf
from   keras.models         import Sequential, Model
from   keras.layers         import Dense, Dropout, Flatten, Activation, BatchNormalization # regularizers
from   keras.layers         import Conv2D, MaxPooling2D
from   keras.callbacks      import ModelCheckpoint, EarlyStopping
from   keras                import backend as K
from   keras.backend        import tensorflow_backend
from   sklearn.metrics      import confusion_matrix
from   pyxvis.io.plots      import show_confusion_matrix, plot_cnn_history
from   pyxvis.io.data       import load_cnn_patches, load_cnn_test_patches

# definition of the CNN architecture using arguments p, d and f:
# there are  n = len(p)  conv2D layers, and m = len(f) fully connected layers
# each conv2D layer has a mask of size p[i] x p[i] x d[i], i=0...n-1
# each conv2D layer has: convolution, ReLU, MaxPooling and Dropout (see details in sequentialCNN in cnn_utils.py)
# each fully connected layer as f[j] nodes, j = 0...m-1
# p             = [5, 3]   # Conv2D mask size
# d             = [4, 6]   # Conv2D channels
# f             = [8]      # fully connected

def CNN(patches_file,type_exec,p,d,f):

    # parameters
    model_file    = 1   # 1 store the best trained model, 0 store the last trained model
    epochs        = 100 # maximal number of epochs in training stage

    # execution type
    if     type_exec == 0: # training and testing
        print('Execution Type 0: Training and testing...')
        do_train      = 1  # 0 means no train, only evaluation,1 means train 
        only_test     = 0  # 1 loads only test data (eg. for sliding windows)
        ev_train      = 1  # accuracy on train data is computed
        ev_test       = 1  # accuracy on test data is computed
        do_conf_mtx   = 1  # compute confusion matrix on testing data
        train_predict = 0  # prediction output of train data is stored in train_predic.npy, 0 = no
        test_predict  = 0  # prediction output of test data is stored in test_predic.npy, 0 = no
        train_layer   = 0  # output layer of train data is stored in train_output.npy, 0 = no
        test_layer    = 0  # output layer of test data is stored in test_output.npy, 0 = no

    elif   type_exec == 1: # eval on testing only
        print('Execution Type 1: Evaluation on testing set only...')
        do_train      = 0  # 0 means no train, only evaluation,1 means train
        only_test     = 0  # 1 loads only test data (eg. for sliding windows)
        ev_train      = 0  # accuracy on train data is computed
        ev_test       = 1  # accuracy on test data is computed
        do_conf_mtx   = 1  # compute confusion matrix on testing data
        train_predict = 0  # prediction output of train data is stored in train_predic.npy, 0 = no
        test_predict  = 0  # prediction output of test data is stored in test_predic.npy, 0 = no
        train_layer   = 0  # output layer of train data is stored in train_output.npy, 0 = no
        test_layer    = 0  # output layer of test data is stored in test_output.npy, 0 = no

    elif   type_exec == 2: # eval on training and testing
        print('Execution Type 2: Evaluation on training and testing set...')
        do_train      = 0  # 0 means no train, only evaluation,1 means train
        only_test     = 0  # 1 loads only test data (eg. for sliding windows)
        ev_train      = 1  # accuracy on train data is computed
        ev_test       = 1  # accuracy on test data is computed
        do_conf_mtx   = 1  # compute confusion matrix on testing data
        train_predict = 0  # prediction output of train data is stored in train_predic.npy, 0 = no
        test_predict  = 0  # prediction output of test data is stored in test_predic.npy, 0 = no
        train_layer   = 0  # output layer of train data is stored in train_output.npy, 0 = no
        test_layer    = 0  # output layer of test data is stored in test_output.npy, 0 = no

    elif   type_exec == 3: # eval layer's output training and testing
        print('Execution Type 3: Computation of layer-output...')
        do_train      = 0  # number of epochs, 0 means no train, only evaluation
        only_test     = 0  # 1 loads only test data (eg. for sliding windows)
        ev_train      = 0  # accuracy on train data is computed
        ev_test       = 0  # accuracy on test data is computed
        do_conf_mtx   = 0  # compute confusion matrix on testing data
        train_predict = 0  # prediction output of train data is stored in train_predic.npy, 0 = no
        test_predict  = 0  # prediction output of test data is stored in test_predic.npy, 0 = no
        train_layer   = 1  # output layer of train data is stored in train_output.npy, 0 = no
        test_layer    = 1  # output layer of test data is stored in test_output.npy, 0 = no

    plot_curves   = 1  # plot loss and accuracy curves
    best_model    = 'cnn_best_model.h5'
    last_model    = 'cnn_last_model.h5'

    #size of parameters
    batch_size    = 128
    droprate      = 0.25

    # tensorflow configuration
    K.set_image_dim_ordering('th')
    print(K.image_data_format())
    config  = tf.ConfigProto(gpu_options=tf.GPUOptions(allow_growth=True))
    session = tf.Session(config=config)
    tensorflow_backend.set_session(session)

    # prepare callbacks
    callbacks    = defineCallBacks(best_model)

    # load patches
    if only_test == 1:
        X_test, Y_test, classes = load_cnn_test_patches(patches_file)
    else:
        X_train, Y_train, X_test, Y_test, classes = load_cnn_patches(patches_file)

    num_classes  = len(classes)
    input_shape  = (X_test.shape[1], X_test.shape[2],X_test.shape[3])

    # CNN architecture defintion
    model        = sequentialCNN(input_shape,droprate,num_classes,p,d,f)

    # training
    if do_train == 1 and only_test == 0:
        deleteWeights(best_model,last_model)
        history = model.fit(X_train, Y_train,
            batch_size      = batch_size,
            epochs          = epochs,
            verbose         = 1,
            validation_data = (X_test, Y_test),
            shuffle         = True,
            callbacks       = callbacks)
        model.save_weights(last_model)
        if plot_curves == 1:
            plot_cnn_history(history)

    # load trained model
    if model_file == 1: # best model
        print('loading best model from '+best_model+' ...')
        model.load_weights(best_model)
    else:               # last model
        print('loading last model from ' +last_model+' ...')
        model.load_weights(last_model)

    # print results in training/testing data
    print('results using '+patches_file+':')
    if ev_train == 1:
        evaluateCNN(model,X_train,Y_train,'training')
    if ev_test == 1:
        evaluateCNN(model,X_test,Y_test,'testing')

    # confusion matrix
    if do_conf_mtx == 1:
        computeconfusionMatrix(model,X_test,Y_test)

    # evaluation of layer outputs
    if test_layer > 0  or test_predict == 1 :
        evaluateLayer(model,K,X_test,'test',test_layer,test_predict)
    if train_layer > 0 or train_predict == 1:
        evaluateLayer(model,K,X_train,'train',train_layer,train_predict)


def sequentialCNN(input_shape,droprate,num_classes,p,d,f):

    n = len(p) # number of conv2D layers
    m = len(f) # number of fc layers

    #Start Neural Network
    model = Sequential()
  
    #convolution 1st layer
    model.add(Conv2D(d[0], kernel_size=(p[0],p[0]), padding="same",activation="relu",input_shape=input_shape))
    model.add(BatchNormalization())
    model.add(MaxPooling2D())
    model.add(Dropout(droprate))
  
    #convolution layer i
    for i in range(1,n):
        # model.add(Conv2D(d[i], kernel_size=(p[i],p[i]), activation="relu",border_mode="same"))
        model.add(Conv2D(d[i], kernel_size=(p[i],p[i]), activation="relu",padding="same"))
        model.add(BatchNormalization())
        model.add(MaxPooling2D())
        model.add(Dropout(droprate))

    #Fully connected layers
    model.add(Flatten())
    for j in range(m):
        model.add(Dense(f[j],use_bias=False))
        model.add(BatchNormalization())
        model.add(Activation('relu'))
        model.add(Dropout(droprate))

    #Fully connected final layer
    model.add(Dense(num_classes))
    model.add(Activation('softmax'))

    model.compile(loss        = keras.losses.categorical_crossentropy,
                  optimizer   = keras.optimizers.RMSprop(),
                  metrics     = ['accuracy'])

    model.summary()

    return model

def deleteWeights(best_model,last_model):
    if os.path.exists(best_model):
        os.remove(best_model)
    if os.path.exists(last_model):
        os.remove(last_model)

def evaluateCNN(model,X,y,st):
    print('evaluating performance in '+st+' set ('+str(y.shape[0])+' samples)...')
    score   = model.evaluate(X,y,verbose=0)
    print(st+' loss:', score[0])
    print(st+' accuracy:', score[1])

def defineCallBacks(model_file):
    callbacks = [
        EarlyStopping(
            monitor        = 'val_acc', 
            patience       = 10,
            mode           = 'max',
            verbose        = 1),
        ModelCheckpoint(model_file,
            monitor        = 'val_acc', 
            save_best_only = True, 
            mode           = 'max',
            verbose        = 0)
    ]
    return callbacks

def sliding_window(image, stepSize, windowSize):
	for i in range(0, image.shape[0], stepSize):
		for j in range(0, image.shape[1], stepSize):
			yield (i, j, image[i:i+windowSize,j:j+windowSize])

def loadSliWinPatches(st_file):
    print('loading image '+ st_file +' for sliding windows...')
    image3 = cv2.imread(st_file)
    image  = image3[:,:,0]
    ROI3   = cv2.imread('ROI.png')
    ROIo   = ROI3[:,:,0]
    ROIo   = ROIo.astype('uint8')
    kernel = np.ones((11,11), np.uint8)
    ROI    = cv2.erode(ROIo, kernel, iterations=1)
    w      = 24
    q      = 12
    X      = np.zeros((100000,1,32,32))
    k      = -1
    for (i, j, window) in sliding_window(image, stepSize=6, windowSize=w):
        if (window.shape[0]==w) and (window.shape[1]==w) and (ROI[i+q,j+q]==1):
            im = cv2.resize(window[:,:], (32,32),interpolation = cv2.INTER_CUBIC)
            k = k+1
            X[k,:,:,:] = im
    print('saving X_test.npy with '+str(k) + ' patches...')
    X_test = X[0:k,:,:,:]/255.0
    Y_test = np.zeros((k,1), dtype=int)
    classes  = [0, 1]
    X_test = X_test.astype('float32')
    np.save('X_test',X_test)
    return X_test, Y_test, classes

def outSliWinPatches(st_file):
    print('computing output image for '+ st_file +' using sliding windows...')
    image3 = cv2.imread(st_file)
    image  = image3[:,:,0]
    ROI3   = cv2.imread('ROI.png')
    ROIo   = ROI3[:,:,0]
    ROIo   = ROIo.astype('uint8')
    kernel = np.ones((11,11), np.uint8)
    ROI = cv2.erode(ROIo, kernel, iterations=1)
    w      = 24
    q      = 12 # w/2
    y_predic = np.load('test_predict.npy')
    n = y_predic.shape[0]
    print(str(n)+' points will be evaluated...')
    image3[0,0,1] = 255 
    image3[0,1,1] = 255 
    image3[1,1,1] = 255 
    image3[1,0,1] = 255 
    t = 0
    k = -1
    for (i, j, window) in sliding_window(image, stepSize=6, windowSize=w):
        if (window.shape[0]==w) and (window.shape[1]==w) and (ROI[i+q,j+q]==1):
            k = k+1
            if k<n:
                if  y_predic[k,1] > 0.95:
                    # print(str(k) + ':' + str(y)+','+str(x)+str(y_predic[k,1]))
                    iq = i+q
                    jq = j+q
                    image3[iq  ,jq  ,0] = 255 
                    image3[iq+1,jq  ,0] = 255 
                    image3[iq  ,jq+1,0] = 255 
                    image3[iq+1,jq+1,0] = 255 
                    image3[iq  ,jq  ,1] = 0 
                    image3[iq+1,jq  ,1] = 0 
                    image3[iq  ,jq+1,1] = 0 
                    image3[iq+1,jq+1,1] = 0 
                    image3[iq  ,jq  ,2] = 0 
                    image3[iq+1,jq  ,2] = 0 
                    image3[iq  ,jq+1,2] = 0 
                    image3[iq+1,jq+1,2] = 0 
                    t = t+1
    print('storing output.png with ' + str(t) +' detected points...')
    cv2.imwrite('output.png',image3)
    return

def evaluateLayer(model,K,X,st,num_layer,test_predict):
    inp        = model.input                                           # input placeholder
    outputs    = [layer.output for layer in model.layers]              # all layer outputs
    functor    = K.function([inp, K.learning_phase()], outputs )       # evaluation function

    test       = X[0]
    test       = test.reshape(1,1,X.shape[2],X.shape[3])
    layer_outs = functor([test, 1.])
    x          = layer_outs[num_layer]
    n          = X.shape[0]
    m          = x.shape[1]

    if test_predict == 1:
        print('computing prediction output in ' +st +' set with '+str(n)+' samples...')
        y = model.predict(X)
        print('saving prediction in '+st+'_predict.npy ...')
        np.save(st+'_predict',y)

    if num_layer>0:
        d = np.zeros((n,m))
        print('computing output layer '+str(num_layer)+ ' in ' +st +' set with '+str(n)+' descriptors of '+str(m)+' elements...')
        for i in range(n):
            test       = X[i]
            test       = test.reshape(1,1,X.shape[2],X.shape[3])
            layer_outs = functor([test, 1.])
            d[i]       = layer_outs[num_layer]
        print('saving layer output in '+st+'_layer_'+str(num_layer)+'.npy ...')
        np.save(st+'_layer_'+str(num_layer),d)


def computeconfusionMatrix(model,X,y):
    Y_prediction = model.predict(X)
    Y_pred_classes = np.argmax(Y_prediction,axis = 1) # classes to one hot vectors 
    Y_true = np.argmax(y,axis = 1)                    # classes to one hot vectors
    show_confusion_matrix(Y_true, Y_pred_classes,'CNN')





"""
"""
__author__ = 'Christian Pieringer'
__created__ = '2019.05.08'

import matplotlib.pyplot as plt

import numpy as np
from cv2 import (resize, INTER_AREA)
from matplotlib.colors import ListedColormap
from matplotlib import animation
from skimage.color import label2rgb
from skimage.measure import label
from skimage.morphology import binary_dilation
from skimage.segmentation import clear_border
from mpl_toolkits.mplot3d import Axes3D
from skimage.feature import plot_matches as _plot_matches


def xvis_sincolormap(w=None, k=None, a=None, th=None):
    """
    This method build the sinmap colormap defined in Appendix B.

    :param w:
    :param k:
    :param a:
    :param th:
    :return:
    """
    # Check input parameters
    if not w:
        w = np.array(
            [1.25 * np.pi / 255, 1.25 * np.pi / 255, 1.25 * np.pi / 255]).T

    if not k:
        k = np.ones((3, 1))

    if not a:
        a = np.zeros((3, 1))

    if not th:
        th = [np.pi / 3, np.pi / 2, -4 * np.pi / 3]

    # Compute the colormap according to Equation (1.5)
    cmap = np.zeros((256, 3))
    x = np.array(range(1, 257))

    cmap[:, 0] = abs(a[0] + k[0] * (np.cos(w[0] * x + th[0])))
    cmap[:, 1] = abs(a[1] + k[1] * (np.cos(w[1] * x + th[1])))
    cmap[:, 2] = abs(a[2] + k[2] * (np.cos(w[2] * x + th[2])))

    # cmap = np.concatenate((cmap, np.ones((256, 1))), axis=1)
    cmap = ListedColormap(cmap)

    return cmap


def show_xray_image(xray_images, color_map='gray'):
    """

    :param xray_images: Numpy array or list of Numpy array with input images
    :param color_map: string with the required colormap. Default is gray.
    :return: None.

    """
    if isinstance(xray_images, np.ndarray):
        fig = plt.figure(figsize=(6, 8))
        ax = fig.add_subplot(1, 1, 1)
        plt.imshow(xray_images, cmap=color_map)
        plt.axis('off')

    elif isinstance(xray_images, list):
        fig = plt.figure(figsize=(12, 8))
        fig.subplots_adjust(hspace=0.4, wspace=0.4)
        for i, img in enumerate(xray_images):
            ax = fig.add_subplot(1, 2, i + 1)
            ax = plt.imshow(img, cmap=color_map)
            plt.axis('off')
    plt.show()


def show_color_array(img, scale=None):
    """
    This function display an array of croped images using
    various colormaps defined previously.

    :params img: the input image in numpy format.

    :returns: None

    """

    # Color definitions
    pyxvix_cmaps = {
        'Gray': 'gray',
        'Jet': 'jet',
        'Hsv': 'hsv',
        'Viridis': 'viridis',
        'Hot': 'hot',
        'Rainbow': 'rainbow',
        'Sinmap': xvis_sincolormap(),
        'Autumn': 'autumn',
        'Spring': 'spring',
    }

    # Generate a copy of the original image.
    img_resized = img.copy()

    # Keep aspect ratio
    if scale:
        width = int(img_resized.shape[1] * scale / 100)
        height = int(img_resized.shape[0] * scale / 100)
        new_size = (width, height)

        # Resize image
        img_resized = resize(img_resized, new_size, interpolation=INTER_AREA)

    ncols, nrows = 3, 3
    fig, fig_axes = plt.subplots(
        ncols=ncols,
        nrows=nrows,
        constrained_layout=False,
        figsize=(16, 8))

    i = 0
    cm = list(pyxvix_cmaps)
    for r in range(nrows):
        for c in range(ncols):
            fig_axes[r, c].imshow(img_resized, cmap=pyxvix_cmaps[cm[i]])
            fig_axes[r, c].axis('off')
            fig_axes[r, c].set_title(cm[i])
            i += 1
    plt.tight_layout(rect=[0, 0.03, 1, 0.90])
    fig.suptitle("Colormap Visualization", fontsize=16)
    plt.show()


def dynamic_colormap(img, cmap, n=128, channels=None):
    """
    Performs the dynamic pseudocolor representations of image 'img'. This will be displayed using different color maps
    as a movie. The first color map is the original ('map'), the next maps will be constructed by circularly shifting
    the the rows of the map in one direction. When a whole shift is done, the direction is changed. This process is
    performed 3 times. Parameter 'channels' defines which columns of the map are shifted. Default is 1:3 (all three
    channels).

    This program can be used for manual inspection. The objects of interest can be more distinguishable when displaying
    the X-ray image as video in different color maps.

    Args:
        img: a numpy array with image
        cmap: a string to define the used colormap.
        n: a integer to define the number of levels in the colormap, default=128.
        channels: a list to define the channels to be changed, default=None.

    Returns: None

    """
    def animate(i, _channels):
        cmap = c['cmap']

        if i >= n:
            cmap = np.vstack([cmap[1:, _channels], cmap[0, _channels][np.newaxis, :]])
        else:
            cmap = np.vstack([cmap[-1, _channels][np.newaxis, :], cmap[0:-1, _channels]])

        new_colormap = ListedColormap(cmap, name='tmp')
        im.set_cmap(new_colormap)
        c['cmap'] = cmap

        return im

    if not channels:
        channels = range(3)

    _cmap = plt.get_cmap(cmap, n)
    _cmap = _cmap(np.linspace(0, 1, n))
    c = {'cmap': _cmap}

    k_img = np.uint8(img.astype(float) / 4)

    fig, ax = plt.subplots(1, 1)
    im = ax.imshow(k_img, cmap=cmap)
    ax.axis('off')

    anim = animation.FuncAnimation(fig, animate, 2*n, fargs=(channels,), interval=1)
    plt.show()


def show_image_as_surface(img, cmap='viridis', elev=-120, azim=40, fsize=None, colorbar=False):
    """
    This function display a mesh plot of an image. This is not a 3D visualization.

    Args:
        img:
        cmap:
        elev:
        azim:

    Returns:
        None
    """
    # Create the x and y coordinate arrays
    xx, yy = np.mgrid[0:img.shape[0], 0:img.shape[1]]

    if not fsize:
        fsize = (8, 8)

    # create the figure
    fig = plt.figure(figsize=fsize)
    ax = fig.gca(projection='3d')
    surf = ax.plot_surface(xx, yy, img,
                           rstride=1,
                           cstride=1,
                           cmap=cmap,
                           linewidth=0)
    ax.view_init(elev, azim)

    if colorbar:
        fig.colorbar(surf, shrink=0.5, aspect=5, pad=0.01)
    ax.axis('off')
    plt.tight_layout()
    plt.show()


def show_series(image_set, series, img_idx, n=4, scale=0.5, _figsize=(18, 8)):
    """

    Args:
        image_set:
        series:
        img_idx:
        n:
        scale:
        _figsize:

    Returns:

    """
    # Load images
    img_list = []

    for i in img_idx:
        img = image_set.load_image(series, i)
        img = resize(img, None, fx=scale, fy=scale, interpolation=INTER_AREA)
        img_list.append(img.copy())

    # Arrange the images creating a new array with a size large enough
    # to contain all the images.
    n_rows = round((len(img_list) / n) + 0.5)
    img_height, img_width = img_list[0].shape
    total_height = n_rows * img_height
    max_width = n * img_width

    final_image = np.zeros((total_height, max_width), dtype=np.uint8)

    # Keep track of where your current image was last placed in the y coordinate
    current_y = 0
    current_x = 0

    for i, image in enumerate(img_list):
        # Add an image to the final array and increment the y coordinate
        h = image.shape[0] + current_y
        w = image.shape[1] + current_x

        final_image[current_y:h, current_x:w] = image.copy()

        current_x += image.shape[1]
        if (i + 1) % n == 0:
            current_y += image.shape[0]
            current_x = 0

    # TODO: add image number on each image
    fig, ax = plt.subplots(1, 1, figsize=_figsize)
    ax.imshow(final_image, cmap='gray')
    ax.axis('off')
    ax.set_title('Series {:s}{:04}'.format(image_set.group_prefix, series))
    plt.show()


def binview(img, mask, color='r', dilate_pixels=1):
    """
    Displays a gray or color image 'img' overlaid by color pixels determined a by binary image 'mask'. It is useful to
    display the edges of an image.

    Args:
        img: gray scale image (X-ray)
        mask: binary image that works as mask
        color: string to define pixel color.
                'r': red (default)
                'g': green
                'b': blue
                'y': yellow
                'c': cyan
                'm': magenta
                'k': black
                'w': white

        dilate_pixels (int): Number of pixels used for dilate the mask.

    Returns:
        img_color (ndarray): output image with a mask overlaid.
    """

    # Defines colors
    # colors = {
    #     'r': np.array([255, 0, 0]),
    #     'g': np.array([0, 255, 0]),
    #     'b': np.array([0, 0, 255]),
    #     'y': np.array([255, 255, 0]),
    #     'c': np.array([0, 255, 255]),
    #     'm': np.array([255, 0, 255]),
    #     'k': np.array([0, 0, 0]),
    #     'w': np.array([255, 255, 255])
    # }
    #
    colors = {
        'r': np.array([1, 0, 0]),
        'g': np.array([0, 1, 0]),
        'b': np.array([0, 0, 1]),
        'y': np.array([1, 1, 0]),
        'c': np.array([0, 1, 1]),
        'm': np.array([1, 0, 1]),
        'k': np.array([0, 0, 0]),
        'w': np.array([1, 1, 1])
    }
    # Create a RGB image from grayscale image.
    img_color = np.dstack((img, img, img))

    # Ensure do not modify the original color image and the mask
    img_color = img_color.copy()

    mask_ = mask.copy()
    # mask_ = dilate(mask_, np.ones((g, g), np.uint8))
    mask_ = binary_dilation(mask_, np.ones((dilate_pixels, dilate_pixels)))

    # Now black-out the area of the mask
    # img_fg = bitwise_and(img, img, mask=mask_)

    # Defines the pixel color used for the mask in the figure.
    cc = colors[color]
    #
    # for i in range(3):
    #     img_color[:, :, i] = cc[i] * img_fg

    # remove artifacts connected to image border
    cleared = clear_border(mask_)
    if np.all(cleared):
        mask_ = cleared

    # label image regions
    label_image = label(mask_)
    img_color = label2rgb(label_image, image=img_color, colors=[cc], bg_label=0)

    return img_color  # add(img_color, img_color)


def h2nh(m):
    """
    Normalize homogeneous coordinates dividing by element of a vector.

    Args:
        m: vector

    Returns:
        The normalized vector (ndarray)

    """

    return (m / m[-1])


def project_edges_on_chessboard(img, mat_p, square_size, nx=10):
    """
    Projects a the real position of the chessboard edges on the calibration pattern.

    Args:
        img (ndarray): the input image
        mat_p (ndarray): projection matrix
        square_size:
        nx:

    Returns:

    """

    x_max = 10
    y_max = 10

    # Generate a meshgrid with real locations edges on the chessboard
    mat_x = np.arange(nx) * square_size
    mat_y = mat_x.copy()
    mat_x, mat_y = np.meshgrid(mat_x, mat_y)

    # Stack all the edges as a tensor where each cell contains the coordinates
    # of the edge in the real world (chess board).
    edges = np.empty(mat_x.shape + (4,))
    edges[:, :, 0] = mat_x
    edges[:, :, 1] = mat_y
    edges[:, :, 2] = np.zeros(mat_x.shape)
    edges[:, :, 3] = np.ones(mat_y.shape)

    # Using Einstein notation to compute and reduce efficiently the matrices.
    me = np.einsum('ijk, mk->ijm', edges, mat_p).reshape(-1, 3).T
    me = h2nh(me)

    # Plot the results
    fontsize = 14
    fig = plt.figure(figsize=(12, 12))
    ax = fig.add_subplot(1, 1, 1)
    ax.imshow(img, cmap='gray')
    ax.scatter(me[0, :], me[1, :], marker='o', edgecolor='orange', facecolor='y')

    # Plot the origin of coordinate in the chessboard system
    orig = np.dot(mat_p, np.array([[0], [0], [0], [1]]))
    orig = h2nh(orig)
    ax.plot(orig[0], orig[1], 'ro')
    ax.text(orig[0], orig[1] - 20, 'O', color='w', fontsize=fontsize, weight='bold')

    orig = np.dot(mat_p, np.array([[0], [250], [0], [1]]))
    orig = h2nh(orig)
    ax.plot(orig[0], orig[1], 'bo')
    ax.text(orig[0], orig[1] - 20, 'Y', color='w', fontsize=fontsize, weight='bold')

    orig = np.dot(mat_p, np.array([[250], [0], [0], [1]]))
    orig = h2nh(orig)
    ax.plot(orig[0], orig[1], 'go')
    ax.text(orig[0], orig[1] - 20, 'X', color='w', fontsize=fontsize, weight='bold')

    ax.axis('off')
    plt.show()


def multivariate_gaussian(pos, mu, sigma):
    """
    Return the multivariate Gaussian distribution on array pos.

    pos is an array constructed by packing the meshed arrays of variables
    x_1, x_2, x_3, ..., x_k into its _last_ dimension.

    [1] https://scipython.com/blog/visualizing-the-bivariate-gaussian-distribution/

    Args:
        pos:
        mu:
        sigma:

    Returns:

    """
    n = mu.shape[0]

    sigma_det = np.linalg.det(sigma)
    sigma_inv = np.linalg.inv(sigma)

    den = np.sqrt(((2 * np.pi) ** n) * sigma_det)

    # This einsum call calculates (x-mu)T.Sigma-1.(x-mu) in a vectorized
    # way across all the input variables.
    fac = np.einsum('...k, kl, ...l->...', pos - mu, sigma_inv, pos - mu)

    return np.exp(-fac / 2) / den


def gaussian_superimposition(img, mat_p, square_size, n_points=30):
    """
    This method projects a Gaussian surface on the calibration pattern.
    """

    x_max = 10
    y_max = 10

    # Generate a meshgrid with real locations edges on the chessboard
    mat_x = np.linspace(0, (x_max - 1) * square_size, n_points)
    mat_y = mat_x.copy()
    mat_x, mat_y = np.meshgrid(mat_x, mat_y)

    # Stack all the edges as a tensor where each cell contains the coordinates
    # of the edge in the real world (chess board).
    edges = np.empty(mat_x.shape + (4,))
    edges[:, :, 0] = mat_x
    edges[:, :, 1] = mat_y
    edges[:, :, 2] = np.zeros(mat_x.shape)
    edges[:, :, 3] = np.ones(mat_y.shape)

    # Mean vector and covariance matrix
    mu = np.array([1125, 1125])
    sigma = np.array([[150000, 0.0], [0.0, 150000]])

    # The distribution on the variables X, Y packed into pos.
    z = multivariate_gaussian(edges[:, :, :2], mu, sigma)
    z = (-1000 * z) / np.max(np.max(z))

    # Add Z coordinates to the edges tensor
    edges[:, :, 2] = z

    me = np.einsum('ijk, mk->ijm', edges, mat_p)

    fig = plt.figure(figsize=(10, 10))
    ax = fig.add_subplot(1, 1, 1)
    ax.imshow(img, cmap='gray')

    # Plot lines on the x-axis first
    for ii in range(me.shape[1]):
        l = me[:, ii]
        l = l / l[:, -1][:, np.newaxis]
        ax.plot(l[:, 0], l[:, 1], '-r')

    # And, finally lines on the y-axis
    for ii in range(me.shape[0]):
        l = me[ii, :]
        l = l / l[:, -1][:, np.newaxis]
        ax.plot(l[:, 0], l[:, 1], '-r')

    ax.axis('off')

    # Plot origin of coordinates
    orig = np.dot(mat_p, np.array([[0], [0], [0], [1]]))
    orig = h2nh(orig)
    ax.plot(orig[0], orig[1], 'ro')
    ax.text(orig[0], orig[1] - 20, 'O', fontsize=14, color='y', weight='bold')

    orig = np.dot(mat_p, np.array([[0], [250], [0], [1]]))
    orig = h2nh(orig)
    ax.plot(orig[0], orig[1], 'bo')
    ax.text(orig[0], orig[1] - 20, 'Y', fontsize=14, color='y', weight='bold')

    orig = np.dot(mat_p, np.array([[250], [0], [0], [1]]))
    orig = h2nh(orig)
    ax.plot(orig[0], orig[1], 'go')
    ax.text(orig[0], orig[1] - 20, 'X', fontsize=14, color='y', weight='bold')

    plt.show()


def plot_bboxes(bounding_boxes, color='lawngreen', linewidth=1.5, ax=None):
    """
    Plot a set of bounding boxes.

    Args:
        bounding_boxes (ndarray): array of bounding boxes (x0, y0, width, height), where x0, y0 is the upper-left corner.
        color (str): Edge color of the bounding box. Use matplotlib color names.
        linewidth (float): Bounding box linewidth.
        ax (object): Figure axes. If None, create a new figure and axes.

    Returns:
        ax (object): Figure axes.
    """
    if not ax:
        fig, ax = plt.subplots(1, 1)

    for b, bbox in enumerate(bounding_boxes):
        rect = plt.Rectangle((bbox[0], bbox[1]), bbox[2], bbox[3], fill=False, edgecolor=color, linewidth=linewidth)
        ax.add_patch(rect)

    return ax


def plot_matches_decorator(_func):
    def wrapper_func(ax, I1, I2, kp1, kp2, matches, **args):
        _func(ax, I1, I2, kp1[:, [1, 0]], kp2[:, [1, 0]], matches, **args)
    return wrapper_func


plot_matches = plot_matches_decorator(_plot_matches)



import numpy               as np
from seaborn                       import heatmap
from sklearn.ensemble              import AdaBoostClassifier
from sklearn.discriminant_analysis import LinearDiscriminantAnalysis
from sklearn.discriminant_analysis import QuadraticDiscriminantAnalysis
from sklearn.neural_network        import MLPClassifier
from sklearn.naive_bayes           import GaussianNB
from sklearn.neighbors             import NearestCentroid
from sklearn.neighbors             import KNeighborsClassifier 
from sklearn.linear_model          import LogisticRegression
from sklearn.ensemble              import RandomForestClassifier
from sklearn.tree                  import DecisionTreeClassifier
from sklearn.svm                   import SVC
from sklearn.metrics               import confusion_matrix, accuracy_score
from sklearn.model_selection       import cross_val_score
from sklearn.base                  import BaseEstimator, ClassifierMixin
from sklearn.neighbors             import KernelDensity
from pyxvis.features.selection     import fse_model, fsel

def define_classifier(bcl):
    if isinstance(bcl,str):
        bcl = clf_model(bcl)
    st = bcl[0]+'('+bcl[1]+')'
    # print('evaluating '+st+'...')
    clf = eval(st)
    return clf

def load_classifiers():
    clfs = [
        'NearestCentroid',               '',
        'KNeighborsClassifier',          'n_neighbors=1',
        'KNeighborsClassifier',          'n_neighbors=3',
        'KNeighborsClassifier',          'n_neighbors=5',
        'KNeighborsClassifier',          'n_neighbors=7',
        'LinearDiscriminantAnalysis',    '',
        'QuadraticDiscriminantAnalysis', '',
        'GaussianNB',                    '',
        'AdaBoostClassifier',            'n_estimators=100',
        'RandomForestClassifier',        'n_estimators=20,random_state = 0',
        'DecisionTreeClassifier',        'max_depth = 4, min_samples_leaf = 8,random_state = 0',
        'SVC',                           'kernel = "linear", gamma=0.2, C=0.1',
        'SVC',                           'kernel = "poly", gamma=0.2, degree = 3, C=0.1',
        'SVC',                           'kernel = "rbf", gamma=0.2,C=0.1',
        'SVC',                           'kernel = "sigmoid", gamma=0.01, C=0.01',
        'LogisticRegression',            'C=0.1,solver="lbfgs"',
        'MLPClassifier',                 'solver="adam", alpha=1e-5,hidden_layer_sizes=(3,2), random_state=1,max_iter=2000',
        'MLPClassifier',                 'solver="adam", alpha=1e-5,hidden_layer_sizes=(10,), random_state=1,max_iter=2000',
        ]
    m = len(clfs)
    bcl = dict()
    bcl['name'] = clfs[0:m:2]
    bcl['parm'] = clfs[1:m:2]
    return bcl

def clf_model(sto,show=0):
    st = sto.lower()
    if st == 'lr':
        name    = 'LogisticRegression'            # name of the classifier
        params  = 'C=0.1,solver="lbfgs"'                              # parameters of the classifier
    elif st == 'dmin':
        name    = 'NearestCentroid'    # name of the classifier
        params  = ''                              # parameters of the classifier
    elif st[0:3] == 'knn':
        name   = 'KNeighborsClassifier'
        params = 'n_neighbors='+st[3:]
    elif st == 'sklda':
        name    = 'LinearDiscriminantAnalysis'    # name of the classifier
        params  = ''                              # parameters of the classifier
    elif st == 'skqda':
        name    = 'QuadraticDiscriminantAnalysis' # name of the classifier
        params  = ''                              # parameters of the classifier
    elif st == 'lda':
        name    = 'LDA'                           # name of the classifier
        params  = ''                              # parameters of the classifier
    elif st == 'qda':
        name    = 'QDA'                           # name of the classifier
        params  = ''                              # parameters of the classifier
    elif st == 'maha':
        name    = 'Mahalanobis'                   # name of the classifier
        params  = 'covi=1'                        # parameters of the classifier
    elif st == 'maha-0':
        name    = 'Mahalanobis'                   # name of the classifier
        params  = 'covi=0'                        # parameters of the classifier
    elif st == 'svm-lin':
        name    = 'SVC'                           # name of the classifier
        params  = 'kernel="linear"'               # parameters of the classifier
    elif st[0:7] == 'svm-rbf':
        name   = 'SVC'
        if len(st)==7:
            C = 1
            gamma = 0.1
        else:
            (gamma,C) = eval(st[7:])
        params = 'kernel = "rbf", gamma='+str(gamma)+',C='+str(C)

    elif st[0:7] == 'svm-pol':
        name   = 'SVC'
        if len(st)==7:
            C = 1
            gamma = 0.1
        else:
            (gamma,C,degree) = eval(st[7:])
        params = 'kernel = "poly", gamma='+str(gamma)+',C='+str(C)+',degree='+str(degree)
        # params = 'kernel = "poly", gamma=0.1, degree = 3, C='+str(C)
    elif st[0:7] == 'svm-sig':
        name   = 'SVC'
        if len(st)==7:
            C = 0.1
            gamma = 0.001
        else:
            (gamma,C) = eval(st[7:])
        params = 'kernel = "sigmoid", gamma='+str(gamma)+',C='+str(C)
    elif st == 'bayes-naive':
        name = 'GaussianNB'
        params = ''
    elif st == 'bayes-kde':
        name = 'KDEClassifier'
        params = 'bandwidth=1.0'
    elif st[0:2] == 'nn':
        name   = 'MLPClassifier'
        if len(st)==2:
            hst = '(10,)'
        else:
            hst = st[2:]
        params = 'solver="adam", alpha=1e-5, random_state=1,max_iter=2000,hidden_layer_sizes='+hst
    elif st == 'rf':
        name = 'RandomForestClassifier'
        params = 'n_estimators=20,random_state = 0'
    elif st == 'tree':
        name = 'DecisionTreeClassifier'
        params = 'max_depth = 4, min_samples_leaf = 8,random_state = 0'
    elif st == 'adaboost':
        name = 'AdaBoostClassifier'
        params = 'n_estimators=100'
    else:
        print('Error: '+st+' does not exist as classification method in clf_model.')
    if show==1:
        print('using classifier: ' + name+'('+params+')...')
    return name,params

def train_classifier(clf,X,d):
    clf = clf.fit(X,d)
    return clf

def test_classifier(clf,Xt):
    ds  = clf.predict(Xt)
    return ds

def train_test_classifier(bcl,X,d,Xt):
    clf = define_classifier(bcl)
    clf = train_classifier(clf,X,d)
    ds  = test_classifier(clf,Xt)
    return ds,clf


class KDEClassifier(BaseEstimator, ClassifierMixin):
    """Bayesian generative classification based on KDE
    from https://jakevdp.github.io/PythonDataScienceHandbook/05.13-kernel-density-estimation.html
    
    Parameters
    ----------
    bandwidth : float
        the kernel bandwidth within each class
    kernel : str
        the kernel name, passed to KernelDensity
    """
    def __init__(self, bandwidth=1.0, kernel='gaussian'):
        self.bandwidth = bandwidth
        self.kernel = kernel
        
    def fit(self, X, y):
        self.classes_ = np.sort(np.unique(y))
        training_sets = [X[y == yi] for yi in self.classes_]
        self.models_ = [KernelDensity(bandwidth=self.bandwidth,
                                      kernel=self.kernel).fit(Xi)
                        for Xi in training_sets]
        self.logpriors_ = [np.log(Xi.shape[0] / X.shape[0])
                           for Xi in training_sets]
        return self
        
    def predict_proba(self, X):
        logprobs = np.array([model.score_samples(X)
                             for model in self.models_]).T
        result = np.exp(logprobs + self.logpriors_)
        return result / result.sum(1, keepdims=True)
        
    def predict(self, X):
        return self.classes_[np.argmax(self.predict_proba(X), 1)]

class Mahalanobis(BaseEstimator, ClassifierMixin):

    def __init__(self, covi=1):
        self.covi = covi # 0 covi = Cte, 1 = covi is different
        
    def fit(self, X, y):
        self.classes  = np.sort(np.unique(y))
        training_sets = [X[y == yi] for yi in self.classes]
        m             = X.shape[1]
        n             = len(self.classes)
        self.mc       = np.zeros((n,m))
        self.Ck       = np.zeros((m,m,n))
        for k in range(n):
            Xk             = training_sets[k]
            self.mc[k]     = np.mean(Xk,axis=0)
            self.Ck[:,:,k] = np.cov(Xk.T)
        if self.covi == 0:
            C       = np.zeros((m,m))
            for k in range(n):
                C  = C + self.Ck[:,:,k]
                C = C/n
            for k in range(n):
                self.Ck[:,:,k] = C
        for k in range(n):
            self.Ck[:,:,k] = np.linalg.inv(self.Ck[:,:,k])
        return self

    def predict_distances(self, Xt):
        n = len(self.classes)
        nt = Xt.shape[0]
        distances = np.zeros((nt,n))
        for k in range(n):
            mk = self.mc[k]
            for i in range(nt):
                xdi  = Xt[i]-mk
                dik   = np.matmul(xdi, self.Ck[:,:,k])
                distances[i,k] = dik.dot(xdi)
        return distances
            
    def predict(self, X):
        return self.classes[np.argmin(self.predict_distances(X), 1)]

class LDA(BaseEstimator, ClassifierMixin):

    def __init__(self, probability=1):
        self.probability = probability

    def fit(self, X, y):
        self.classes  = np.sort(np.unique(y))
        training_sets = [X[y == yi] for yi in self.classes]
        N             = X.shape[0]
        m             = X.shape[1]
        n             = len(self.classes)
        self.mc       = np.zeros((n,m))
        Cw            = np.zeros((m,m))
        pk            = np.zeros((n,1))
        for k in range(n):
            Xk    = training_sets[k]
            Lk    = Xk.shape[0]
            pk[k] = Lk/N
            C     = np.linalg.inv(np.cov(Xk.T))
            Cw    = Cw + C*(Lk-1)
            self.mc[k] = np.mean(Xk,axis=0)
        Cw = Cw/(N-n)
        self.invCw = np.linalg.inv(Cw)
        if self.probability == 1:
            self.logp = np.log(np.ones((n,1))/n)
        elif self.probability == 0:
            self.logp = np.log(pk)
        else:
            self.logp = np.log(self.probability)
        return self

    def predict_distances(self, Xt):
        n = len(self.classes)
        nt = Xt.shape[0]
        distances = np.zeros((nt,n))
        C  = self.invCw
        for k in range(n):
            mk = self.mc[k]
            for i in range(nt):
                xdi  = Xt[i]-mk
                dik   = np.matmul(xdi, C)
                distances[i,k] = -0.5*dik.dot(xdi)+self.logp[k]
        return distances
            
    def predict(self, X):
        return self.classes[np.argmax(self.predict_distances(X), 1)]

class QDA(BaseEstimator, ClassifierMixin):

    def __init__(self, probability=1):
        self.probability = probability

    def fit(self, X, y):
        self.classes  = np.sort(np.unique(y))
        training_sets = [X[y == yi] for yi in self.classes]
        N             = X.shape[0]
        m             = X.shape[1]
        n             = len(self.classes)
        self.Ck       = np.zeros((m,m,n))
        self.mc       = np.zeros((n,m))
        pk            = np.zeros((n,1))
        for k in range(n):
            Xk             = training_sets[k]
            Lk             = Xk.shape[0]
            pk[k]          = Lk/N
            self.mc[k]     = np.mean(Xk,axis=0)
            self.Ck[:,:,k] = np.linalg.inv(np.cov(Xk.T))
        if self.probability == 1:
            self.logp = np.log(np.ones((n,1))/n)
        elif self.probability == 0:
            self.logp = np.log(pk)
        else:
            self.logp = np.log(self.probability)
        return self

    def predict_distances(self, Xt):
        n = len(self.classes)
        nt = Xt.shape[0]
        distances = np.zeros((nt,n))
        for k in range(n):
            C  = self.Ck[:,:,k]
            Cp = -0.5*np.log(np.linalg.det(C))+self.logp[k]
            mk = self.mc[k]
            for i in range(nt):
                xdi  = Xt[i]-mk
                dik   = np.matmul(xdi, C)
                distances[i,k] = -0.5*dik.dot(xdi) + Cp
        return distances
            
    def predict(self, X):
        return self.classes[np.argmax(self.predict_distances(X), 1)]

def nn_definition(n,N):
    W  = [None]
    b  = [None]
    m1 = len(n)
    for k in range(1,m1):
        W    = W  + [np.random.rand(n[k],n[k-1])]
        b    = b  + [np.random.rand(n[k],1)]
    return W,b

def nn_forward_propagation(X,W,b):
    a = [None]
    a[0]  = X
    m1 = len(W)
    for k in range(1,m1):
        zk = W[k].dot(a[k-1])+b[k]
        a = a + [1/(1+np.exp(-zk))]
    return a

def nn_backward_propagation(Y,a,W,b):
    m = len(W)-1
    N = Y.shape[1]
    dam = a[m]-Y
    dW = [None]
    db = [None]

    # Derivatives
    for k in range(1,m+1):
        dW   = dW + [np.zeros([W[k].shape[0],W[k].shape[1]])]
        db   = db + [np.zeros([b[k].shape[0],1])]
    for k in range(m,0,-1):
        if k == m:
            dak = dam
        ds    = np.multiply(a[k], 1-a[k]) 
        Gk    = np.multiply(dak,ds)
        dW[k] = np.matmul(Gk,a[k-1].transpose())/N
        db[k] = (np.sum(Gk,axis=1,keepdims=True))/N 
        dak   = np.matmul(W[k].transpose(),Gk)
    
    return dW,db

def nn_parameters_update(W,b,dW,db,alpha):
    m1 = len(W)

    # Updates
    for k in range(1,m1):
        b[k] = b[k] - alpha*db[k]
        W[k] = W[k] - alpha*dW[k]

    return W,b

def nn_loss_function(a,Y):
    m = len(a)-1
    N = Y.shape[1]
    Ys = a[m]
    dam = Ys-Y
    d2 = np.multiply(dam,dam)
    ds = np.sqrt(np.sum(d2,axis=0,keepdims=True))/N
    loss = np.sum(ds)
    return loss




"""
This module contains all the methods used for image processing.

"""

__author__ = 'Christian Pieringer'
__created__ = '2019.05.26'

import numpy as np
from cv2 import filter2D, CV_64F
from cv2 import GaussianBlur, medianBlur
from skimage.feature import canny
from scipy.signal import convolve2d

from .helpers import kfunctions as kfunc
# from pyxvis.processing.helpers import kfunctions as kfunc


def dual_energy(img_1, img_2, lut):
    """
    Allows for dual energy image computation.

    Args:
        img_1 (ndarray): first image
        img_2 (ndarray): second image
        lut (ndarray): the lookup table use during computation

    Raises:
        TypeError: invalid input type for images

    Returns:
        _dual_energy (numpy array): the dual energy image.

    """
    # Check input parameters
    if not (isinstance(img_1, np.ndarray) and isinstance(img_2, np.ndarray)):
        raise TypeError('{}/{} are not an accepted input type for img_1 or img_2'.format(type(img_1), type(img_2)))

    nrows, ncols = img_1.shape
    _dual_energy = np.zeros((nrows, ncols))

    for i in range(nrows):
        for j in range(ncols):
            _dual_energy[i, j] = lut[img_2[i, j], img_1[i, j]]

    return _dual_energy


def shading(mat_x, mat_r1, mat_r2, i1, i2):
    """
    This method performs a shading correction. There are two reference images: 
    image R1, of a thin plate, image r2, of a thick plate. The gray values i1 
    and i2 are the ideal gray values for the first and second image, respectively.
    From r1, r2, i1 and i2, offset and gain, correction matrices A and B are 
    calculated assuming a linear transformation between the original X-ray 
    image X and corrected X-ray image y.
    """

    mat_a = (i2 - i1) / (mat_r2 - mat_r1)
    mat_b = i1 - (mat_r1 * mat_a)
    mat_y = (mat_a * mat_x) + mat_b

    return mat_y


def fspecial(kernel_name, size=3, sigma=1.0):
    """
    This method mimic the 'fspecial' gaussian MATLAB function to create 
    special kernel functions used in computer vision.

    Reference: https://src-bin.com/en/q/1064ef9

    :param kernel_name: a string to define the type of window:
                'sobel': Sobel kernel
                'prewitt': Prewitt kernel
                'gaussian': standard Gaussian window
                'log': Laplacian of Gaussian Kernel

    :param size: integer to define the length of the window
    :param sigma: a float number to define the sigma of the kernel
    """

    _kernel_kwarg = {
        'sobel': kfunc.sobel_kernel,
        'prewitt': kfunc.prewitt_kernel,
        'gaussian': kfunc.gaussian_kernel,
        'log': kfunc.log_kernel
    }

    # _img_generate kernel function
    kernel = _kernel_kwarg[kernel_name](size, sigma)

    return kernel


# def bwareaopen(img, area, connectivity=4):
#     """
#     This method emulate the behavior of the Matlab(c) 'bwareaopen'.
#
#     :param img: a binary ndarray that holds
#     :param area:
#     :param connectivity:
#
#     :returns: a binary ndarray with the result
#
#     """
#     num_labels, labels, stats, centroids = connectedComponentsWithStats(
#         img.astype(np.uint8), connectivity)
#     bw_labels = np.zeros_like(labels, dtype=np.uint8)
#
#     # Loop throught labels and remove areas according to the size
#     #     areas = stats[1:, cv.CC_STAT_AREA]  #[s[4] for s in stats]
#     #     print(areas[areas > area])
#     #     sorted_idx = np.argsort(areas)
#
#     for idx in range(1, num_labels - 1):
#         x, y, w, h, size = stats[idx, :]
#
#         if size > area:
#             bw_labels[labels == idx] = 1
#
#     return bw_labels


def im_grad(img, kernel):
    """
    This method compute the gradient of a image 'img'. It uses the OpenCV
    function "filter2D". In the filter function 'same' is the default and the
    only one option.

    :param img: a uint8 input image.
    :param kernel: a kernel as nd-array. Commonly used kernels are Prewitt,
                    _img_gaussians, or Sobel.

    :returns: The images of magnitude (mag) and angle (angle).

    """

    # TODO: check arguments

    gradi = filter2D(img.astype('double'), CV_64F, kernel)
    gradj = filter2D(img.astype('double'), CV_64F, kernel.T)

    mag = np.sqrt(gradi ** 2 + gradj ** 2)
    angle = np.arctan2(gradj, gradi)

    return mag, angle


def linimg(img, t=255):
    """
    This method computes the Linear enhancement of image I from 0 to 255,
    or 0 to 1.

    :param img: a grayvalue image I.
    :param t: if t==255 output will be a unit8 image with gray values going from
              0 to 255. Else, output will a double image with gray values from
              0 to t.

    :returns: enhanced image J (uint8 or double) so that:
                J = m*I + b; where min(J) = 0, max(J) = 255 or 1.
    
    """

    img = np.double(img.copy())
    mi = np.min(img.flatten())
    mx = np.max(img.flatten())
    d = mx - mi

    if d == 0:
        out = (img / mx) * (t / 2)
    else:
        out = ((img - mi) / d) * t

    if t == 255:
        out = np.round(out)

    return out.astype('uint8')


def im_median(_img, k):
    """
    Applies median filter to an gray-scaled image.

    Args:
    _img (ndarray): a gray-scale image.
    k (int): value to define number of pixels of the median mask
    
    Returns:
     img_j (ndarray): filtered image.
    """
    img_j = medianBlur(_img, k)

    return img_j


def im_gaussian(_img, k, sigma=None):
    """
    This method applies a gaussian filtering on image _img.

    Args:
        _img (ndarray): a gray-valued image
        k (int): value to define the mask size
        sigma (float): value to control the sigma parameter. Default k/8.5.
    
    Returns:
        img_j (ndarray): filtered image
    """
    if not sigma:
        sigma = k / 8.5

    img_j = GaussianBlur(_img, (k, k), sigma)

    return img_j


def im_equalization(img, gamma=1.0):
    """
    TODO: write documentation.
    """

    img_x = img.copy()
    img_x = 255 * ((img_x - img_x.min()) / (img_x.max() - img_x.min())) ** gamma

    return img_x


def im_average(input_img, k):
    """
    It performs a linear average filtering of "input_img".
    
    Args:
        img (array): a grayvalue image.
        k (int): size of the average window  

    Returns:
        out_img (array): filtered image so that out_img is convolution of "input_img" with an average 
                         mask h = ones(k,k)/k^2.
    """
    if not isinstance(input_img, np.ndarray):
        raise TypeError('Function expects an ndarray for input image but received {}.'.format(type(input_img)))

    if not isinstance(k, int):
        raise TypeError('Function expects an ndarray for k but received {}.'.format(type(k)))

    h = np.ones((k, k)) / (k ** 2)
    out_img = convolve2d(np.double(input_img), h, 'same')

    return out_img


def hist_forceuni(img, show=False):
    """
    TODO: write documentation.
    """

    img_x = im_equalization(img)
    n, m = img_x.shape
    y = np.zeros((n * m, 1), dtype=np.uint8)
    j = np.argsort(img_x.flatten())
    z = np.zeros((n * m, 1), dtype=np.uint8)
    d = np.int(np.fix((n * m / 256) + 0.5))

    for i in range(255):
        z[i * d:(i + 1) * d] = i * np.ones((d, 1))  # , dtype=np.uint8)
    z[255 * d:n * m] = 255 * np.ones((n * m - 255 * d, 1))  # , dtype=np.uint8)

    y[j] = z
    y = y.reshape(n, m)

    return y


class Edge:
    """
    This class implements methods for edge detection required for examples used 
    in the book.
    """

    def __init__(self, _method, _thres=None, _sigma=1.0):
        """
        Args:
            method (str): It defines the detector method:
                'log': Laplacian of Gaussian
                'canny': Canny edge detection
            thres (float): It defines the threshold to noise supression
            sigma (float): It specifies the scale of the kernel
        """

        _valid_methods = ['log', 'canny', 'sobel']

        super().__init__()

        # Check input parameters
        if _method not in _valid_methods:
            raise ValueError('{} is not a valid method for edge detection.'.format(_method))

        self.method = _method
        self.thres = _thres
        self.sigma = _sigma
        self._returns = None  # Keep track outputs

    def detect_zero_crossing(self, edges):
        """
        Performs zero crossing detection.
        
        Args:
            edges (float ndarray): input image in edges
    
        Return:
            out_img (ndarray): image of edges.
        """

        # def any_neighbor_zero(img, i, j):
        #     """
        #     This function verifies neighborhood around boundary pixels.
        #
        #     """
        #     for k in range(-1, 2):
        #         for l in range(-1, 2):
        #             if img[i + k, j + k] == 0:
        #                 return True
        #     return False
        #
        # # Makes a copy of the original image and
        # # marks as 1 pixels greater that 0.
        # _edges = edges.copy()
        # _edges[_edges > self.thres] = 1
        # _edges[_edges <= self.thres] = 0
        #
        # # _edges = np.abs(_edges)
        # # _edges = _edges > self.thres
        #
        # # print(np.max(_edges), np.min(_edges))
        #
        # out_img = np.zeros(_edges.shape)
        #
        # for i in range(1, _edges.shape[0] - 1):
        #     for j in range(1, _edges.shape[1] - 1):
        #         if _edges[i, j] > 0 and any_neighbor_zero(_edges, i, j):
        #             out_img[i, j] = 1

        # Check zero crossing four directions and aggregate the output.
        out = np.diff(np.sign(edges), prepend=np.ones((edges.shape[0], 1)), axis=1) != 0
        out = np.bitwise_or(out, np.diff(np.sign(edges), prepend=np.ones((1, edges.shape[1])), axis=0) != 0)
        out = np.bitwise_or(out, (np.diff(np.sign(edges)[:, ::-1], prepend=np.ones((1, edges.shape[1])), axis=0) != 0)[:, ::-1])
        out = np.bitwise_or(out, (np.diff(np.sign(edges)[::-1, :], prepend=np.ones((1, edges.shape[1])), axis=0) != 0)[::-1, :])
        # Check gradients is on the threshold.

        grads_diff = np.abs(np.diff(edges, prepend=np.ones((edges.shape[0], 1)), axis=1)) > self.thres
        grads_diff = np.bitwise_or(grads_diff, np.abs(np.diff(edges, prepend=np.ones((1, edges.shape[0])), axis=0)) > self.thres)
        # grads_diff = np.bitwise_or(grads_diff, (np.abs(
        #     np.diff(edges[:, ::-1], prepend=np.ones((edges.shape[0], 1)),
        #             axis=1)) > self.thres)[:, ::-1])
        # grads_diff = np.bitwise_or(grads_diff, (np.abs(
        #     np.diff(edges[::-1, :], prepend=np.ones((1, edges.shape[0])),
        #             axis=0)) > self.thres)[::-1, :])

        out_img = np.bitwise_and(out, grads_diff)

        return out_img

    def fit(self, img):
        """
        This method detect edges in an image. Method must be choose during class
        instantiation.

        Args:
            img: input image in numpy array format.

        Returns:
            None
        """

        self._returns = dict()
        self._returns.update({'edges': None})

        if self.method == 'log':
            _ksize = int(2 * np.ceil(3 * self.sigma) + 1)
            _log_kernel = fspecial(self.method, _ksize, self.sigma)
            _log = filter2D(img.astype('double'), CV_64F, _log_kernel)
            _log = _log / np.sum(_log.flatten())  # Normalize gradients to sum 0
            _edges = self.detect_zero_crossing(_log)

            self._returns.update({'kernel': _log_kernel})  # Keep used kernel

        elif self.method == 'canny':
            """
            Canny edge detector. This option implement a wrapper to apply the 
            scikit image implementation for this edge detector. It present an 
            interface with similar parameters to the MATLAB version.    
            """
            _edges = canny(
                img.astype('double'),
                sigma=self.sigma,
                low_threshold=self.thres,
                high_threshold=self.thres / 0.4
            )

        # Keep the output if does exists
        self._returns.update({'edges': _edges})

    @property
    def edges(self):
        return self._returns['edges']

    @property
    def kernel(self):
        return self._returns['kernel'] if self.method != 'canny' else None


def res_minio(_img_g, h, method='minio'):
    """

    Args:
        _img_g: blured image as ndarray.
        h: PSF as ndarray.
        method: integer value to select minimization method:
                'minio': ||f_n - g|| -> min
                'other': ||f|| -> min

    Returns:
        _img_f: the restored image as ndarray.
    
    Raises:
        ValueError: if h is not ndarray.
        Exception
    """

    _valid_methods = ['minio', 'other']

    # Check input parameters
    # Input image _img_g
    if not isinstance(_img_g, np.ndarray):
        raise Exception('{} is not valid input format for the blured image _img_g.'.format(type(_img_g)))

    # PSF h
    if not isinstance(h, np.ndarray):
        raise Exception('{} is not a valid format for PSF.'.format(type(h)))

    # Method
    if method not in _valid_methods:
        raise ValueError('method do not meet {} as a valid method'.format(method))

    # Main
    nh = h.shape[-1]
    n = _img_g.shape[-1]
    m = n + nh - 1
    _mat_h = np.zeros((n, m))

    # Composing H matrix
    for i in range(n):
        _mat_h[i, i:i + nh] = h.ravel()

    _lambda = 1e6  # Lagrange multiplier

    if method == 'minio':
        _mat_p = np.hstack([np.eye(n, dtype='float'), np.zeros((n, nh - 1))])
        _img_f = _img_g @ (_lambda * _mat_h + _mat_p)
        _img_f = _img_f @ (np.linalg.inv((_lambda * _mat_h).T @ _mat_h + _mat_p.T @ _mat_p)).T
    else:
        _img_f = _img_g @ _mat_h
        _img_f = _img_f @ (np.linalg.inv((_lambda * _mat_h).T @ _mat_h + np.eye(m, dtype='float'))).T

    # Clip the image between 0 and 255
    _img_f[_img_f < 0] = 0
    _img_f[_img_f > 255] = 255

    return _img_f

# Added by D.Mery on June, 2nd, 2020

from scipy.ndimage.filters import convolve

def conv2(x, y, mode='same'):
    """
    Emulate the function conv2 from Mathworks.

    Args:
        x: 
    Usage:

    z = conv2(x,y,mode='same')

    TODO: 
     - Support other modes than 'same' (see conv2.m)
    """

    if not(mode == 'same'):
        raise Exception("Mode not supported")

    # Add singleton dimensions
    if (len(x.shape) < len(y.shape)):
        dim = x.shape
        for i in range(len(x.shape),len(y.shape)):
            dim = (1,) + dim
        x = x.reshape(dim)
    elif (len(y.shape) < len(x.shape)):
        dim = y.shape
        for i in range(len(y.shape),len(x.shape)):
            dim = (1,) + dim
        y = y.reshape(dim)

    origin = ()

    # Apparently, the origin must be set in a special way to reproduce the results of scipy.signal.convolve and Matlab.
    for i in range(len(x.shape)):
        if ( (x.shape[i] - y.shape[i]) % 2 == 0 and
             x.shape[i] > 1 and
             y.shape[i] > 1):
            origin = origin + (-1,)
        else:
            origin = origin + (0,)

    z = convolve(x, y, mode='constant', origin=origin)

    return z


def gradlog(img, sigma, th):
    """
    Edge detection using LoG and gradient.

    Args:
        img (ndarray): input image
        sigma (float): sigma parameters for Edge instance
        th (float): threshold

    Returns:
        edge (ndarray): edge detection. It is or(E_log, E_grad), where E_log is edge detection using LoG, and E_grad is
                        edge detection using gradient of I thresholding by th. Gradient of I is computed after Gaussian 
                        low pass filtering (using sigma).
    """
    detector = Edge('log', 1e-10, sigma)
    detector.fit(img)
    elog = detector.edges
    hsize = np.round((8.5 * sigma) / 2) * 2 + 1
    g = kfunc.gaussian_kernel(hsize, sigma)
    kernel = np.array([-1.0, 1.0])
    h = conv2(g, kernel)
    grad, _ = im_grad(img,h)
    egrad = grad > th
    edge = egrad | elog

    return edge

# vim: set ts=4 sts=4 sw=4 expandtab smartindent:
#
# This file was originally part of the octave-forge project
# Ported to python by Luis Pedro Coelho <luis@luispedro.org> (February 2008)
# Copyright (C) 2006       Soren Hauberg
# Copyright (C) 2008-2010  Luis Pedro Coelho (Python port)
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details. 
# 
# You should have received a copy of the GNU General Public License
# along with this file.  If not, see <http://www.gnu.org/licenses/>.

# changed by skimage.segmentation.find_boundaries
#def bwperim(bw, n=4):
#    """
#    Find the perimeter of objects in binary images. A pixel is part of an object perimeter if its value is one and 
#    there is at least one zero-valued pixel in its neighborhood. By default the neighborhood of a pixel is 4 nearest 
#    pixels, but if `n` is set to 8 the 8 nearest pixels will be considered.
#
#    Args:
#        bw (ndarray): the input binary image
#        n (int): Connectivity. Must be 4 or 8 (default: 4)
#    
#    Returns:
#        perim (bool): A binary image with the perimeter of the input mask.
#
#    Raises:
#        ValueError: Not valid values for n. Method requires that n be 4 or 8.
#    """
#    if n not in (4, 8):
#        raise ValueError('bwperim: Not valid value for n. Method expect n equals to 4 or 8.')
#
#    rows,cols = bw.shape
#
#    # Translate image by one pixel in all directions
#    north = np.zeros((rows,cols))
#    south = np.zeros((rows,cols))
#    west = np.zeros((rows,cols))
#    east = np.zeros((rows,cols))
#
#    north[:-1,:] = bw[1:,:]
#    south[1:,:]  = bw[:-1,:]
#    west[:,:-1]  = bw[:,1:]
#    east[:,1:]   = bw[:,:-1]
#    idx = (north == bw) & \
#          (south == bw) & \
#          (west  == bw) & \
#          (east  == bw)
#
#    if n == 8:
#        north_east = np.zeros((rows, cols))
#        north_west = np.zeros((rows, cols))
#        south_east = np.zeros((rows, cols))
#        south_west = np.zeros((rows, cols))
#        north_east[:-1, 1:]   = bw[1:, :-1]
#        north_west[:-1, :-1] = bw[1:, 1:]
#        south_east[1:, 1:]     = bw[:-1, :-1]
#        south_west[1:, :-1]   = bw[:-1, 1:]
#        idx &= (north_east == bw) & \
#               (south_east == bw) & \
#               (south_west == bw) & \
#               (north_west == bw)
#
#    return ~idx * bw
#


# Pre-trained models
# This code can extract the embeddings of many pre-trained models. 
# The idea is to extract one embedding for a set of images 
# 
# variables: model_id output_layer
# 
# model_id can be (in parenthesis the available output_layer)
# 
#      0 : ResNet50 (0,1,2)
#      1 : VGG16 (0,1,2)        
#      2 : VGG19 (0,1,2)
#      3 : InceptionV3 (0)
#      4 : InceptionResNetV2 (0)
#      5 : Xception (0)
#      6 : MobileNet (1,2)
#      7 : SqueezeNet (1,2)        
#      8 : AlexNet (1,2)        
#      9 : GoogleNet (1,2)        
#     10 : ShuffleNet (1,2)        
#     11 : DenseNet121 (1,2)        
#     12 : ZfNet512 (1,2)        
#     13 : RCNN_ILSVRC13 (1,2)     
# 
# output_layer can be
# 
#      0 : Keras, layer before softmax
#      1 : ONNX,  last layer (after softmax) with 1000 features
#      2 : ONNX,  layer before softmax 
# 
# Example 1: Extraction of ResNet50 using Keras
# variables: model_id = 0, output_layer = 0
# 
# Example 2: Extraction of AlexNet using ONNX, layer before softmax 
# variables: model_id = 8, output_layer = 2

import os, fnmatch
import numpy as np
import mxnet as mx
import matplotlib.pyplot as plt

from mxnet.gluon.data.vision import transforms
from mxnet.contrib.onnx.onnx2mx.import_model import import_model


from keras.applications.resnet50 import ResNet50
from keras.applications.resnet50 import preprocess_input as proc0

from keras.applications.vgg16 import VGG16
from keras.applications.vgg16 import preprocess_input as proc1

from keras.applications.vgg19 import VGG19
from keras.applications.vgg19 import preprocess_input as proc2

from keras.applications.inception_v3 import InceptionV3
from keras.applications.inception_v3 import preprocess_input as proc3

#from keras.applications.inception_resnet_v2 import InceptionResNetV2
#from keras.applications.inception_resnet_v2 import preprocess_input as proc4

from keras.applications.xception import Xception
from keras.applications.xception import preprocess_input as proc5

from keras.preprocessing import image

from sklearn.metrics import confusion_matrix

from pyxvis.io.misc import dirfiles,num2fixstr


def prt_model(model_id,output_layer):
    # ResNet50

    print('loading model...')
    # 0 ResNet50
    if model_id == 0: 
        if output_layer == 0: # Keras
            model_name = 'ResNet50_Keras'
            model = ResNet50(weights='imagenet', include_top=False)
        img_size = 224
        #img_size = 30
        model_path = '../models/resnet50.onnx'
        datain = 'gpu_0/data_0'
        if output_layer == 1: # ONNX - Last Layer
            model_name = 'ResNet50_ONNX_L'
            layerout = 'last'
        elif output_layer == 2: # ONNX - Previous Layer
            model_name = 'ResNet50_ONNX_P'
            layerout = 'flatten0_output'
    
    # 1 VGG16        
    elif model_id == 1: 
        if output_layer == 0: # Keras
            model_name = 'VGG16_Keras'
            model = VGG16(weights='imagenet', include_top=False)
        #img_size = 224
        img_size = 60
        model_path = '../models/vgg16.onnx'
        datain = 'data'
        if output_layer == 1: # ONNX - Last Layer
            model_name = 'VGG16_ONNX_L'
            layerout = 'last'
        elif output_layer == 2: # ONNX - Previous Layer
            model_name = 'VGG16_ONNX_P'
            layerout = 'flatten2_output'

    # 2 VGG19
    elif model_id == 2: 
        if output_layer == 0: # Keras
            model_name = 'VGG19_Keras'
            model = VGG19(weights='imagenet', include_top=False)
        #img_size = 224
        img_size = 60
        model_path = '../models/vgg19.onnx'
        datain = 'data_0'
        if output_layer == 1: # ONNX - Last Layer
            model_name = 'VGG19_ONNX_L'
            layerout = 'last'
        elif output_layer == 2: # ONNX - Previous Layer
            model_name = 'VGG19_ONNX_P'
            layerout = 'flatten2_output'

    # 3 InceptionV3
    elif model_id == 3: 
        if output_layer == 0: # Keras
            model_name = 'InceptionV3_Keras'
            model = InceptionV3(weights='imagenet', include_top=False)
        #img_size = 224
        img_size = 90

    # 4 InceptionResNetV2
    #elif model_id == 4: 
    #    if output_layer == 0: # Keras
    #        model_name = 'InceptionResNetV2_Keras'
    #        model = InceptionResNetV2(weights='imagenet', include_top=False)
    #    #img_size = 224
    #    img_size = 90

    # 5 Xception
    elif model_id == 5: 
        if output_layer == 0: # Keras
            model_name = 'Xception_Keras'
            model = Xception(weights='imagenet', include_top=False)
        #img_size = 224
        img_size = 30
    
    # 6 MobileNet
    elif model_id == 6: 
        model_path = '../models/mobilenet.onnx'
        datain = 'data'
        if output_layer == 1: # ONNX - Last Layer
            model_name = 'MobileNet_ONNX_L'
            layerout = 'last'
        elif output_layer == 2: # ONNX - Previous Layer
            model_name = 'MobileNet_ONNX_P'
            layerout = 'mobilenetv20_features_pool0_fwd_output' # not good

    # 7 SqueezeNet        
    elif model_id == 7:
        model_path = '../models/squeezenet.onnx'
        datain = 'data'
        if output_layer == 1: # ONNX - Last Layer
            model_name = 'SqueezeNet_ONNX_L'
            layerout = 'last'
        elif output_layer == 2: # ONNX - Previous Layer
            model_name = 'SqueezeNet_ONNX_P'
            layerout = 'pooling3_output' # not good

    # 8 AlexNet        
    elif model_id == 8:
        model_path = '../models/alexnet.onnx'
        datain = 'data_0'
        if output_layer == 1: # ONNX - Last Layer
            model_name = 'AlexNet_ONNX_L'
            layerout = 'last'
        elif output_layer == 2: # ONNX - Previous Layer
            model_name = 'AlexNet_ONNX_P'
            layerout = 'flatten2_output'

    # 9 GoogleNet        
    elif model_id == 9:
        model_path = '../models/googlenet.onnx'
        datain = 'data_0'
        if output_layer == 1: # ONNX - Last Layer
            model_name = 'GoogleNet_ONNX_L'
            layerout = 'last'
        elif output_layer == 2: # ONNX - Previous Layer
            model_name = 'GoogleNet_ONNX_P'
            layerout = 'flatten0_output' # not good

    # 10 ShuffleNet        
    elif model_id == 10:
        model_path = '../models/shufflenet.onnx'
        datain = 'gpu_0/data_0'
        if output_layer == 1: # ONNX - Last Layer
            model_name = 'ShuffleNet_ONNX_L'
            layerout = 'last'
        elif output_layer == 2: # ONNX - Previous Layer
            model_name = 'ShuffleNet_ONNX_P'
            layerout = 'flatten0_output' # not good

    # 11 DenseNet121        
    elif model_id == 11:
        model_path = '../models/densenet121.onnx'
        datain = 'data_0'
        if output_layer == 1: # ONNX - Last Layer
            model_name = 'DenseNet121_ONNX_L'
            layerout = 'last'
        elif output_layer == 2: # ONNX - Previous Layer
            model_name = 'DenseNet121_ONNX_P'
            layerout = 'pad124_output' # not good

    # 12 ZfNet512        
    elif model_id == 12:
        model_path = '../models/zfnet512.onnx'
        datain = 'gpu_0/data_0'
        if output_layer == 1: # ONNX - Last Layer
            model_name = 'ZfNet512_ONNX_L'
            layerout = 'last'
        elif output_layer == 2: # ONNX - Previous Layer
            model_name = 'ZfNet512_ONNX_P'
            layerout = 'flatten2_output' # not good

    # 13 RCNN_ILSVRC13     
    elif model_id == 13:
        model_path = '../models/rcnn_ilsvrc13.onnx'
        datain = 'data_0'
        if output_layer == 1: # ONNX - Last Layer
            model_name = 'RCNN_ILSVRC13_ONNX_L'
            layerout = 'last'
        elif output_layer == 2: # ONNX - Previous Layer
            model_name = 'RCNN_ILSVRC13_ONNX_P'
            layerout = 'flatten2_output' # not good



    if output_layer > 0:  
        print('loading model '+model_path+'...')
        sym, arg_params, aux_params = import_model(model_path)
        if len(mx.test_utils.list_gpus())==0:
            ctx = mx.cpu()
        else:
            ctx = mx.gpu(0)
        
        all_layers = sym.get_internals()
        print(all_layers.list_outputs())
        if layerout == 'last':
            sym3 = sym
        else:
            sym3 = all_layers[layerout]
        model = mx.mod.Module(symbol=sym3, context=ctx, label_names=None, data_names=[datain] )
        image_size = (224,224)
        img_size = image_size[1]
        model.bind(data_shapes=[(datain, (1, 3, image_size[0], image_size[1]))])
        model.set_params(arg_params, aux_params)

    print('model '+model_name+' loaded.')
    return model,img_size,model_name

def get_image(path, show=False):
    img = mx.image.imread(path)
    if img is None:
        return None
    if show:
        plt.imshow(img.asnumpy())
        plt.axis('off')
    return img

def preprocess(img):   
    transform_fn = transforms.Compose([
    transforms.Resize(256),
    transforms.CenterCrop(224),
    transforms.ToTensor(),
    transforms.Normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
    ])
    img = transform_fn(img)
    img = img.expand_dims(axis=0)
    return img

def get_features(mod,path):
    img = get_image(path, show=True)
    img = preprocess(img)
    input_blob = img
    data = mx.nd.array(input_blob)
    db = mx.io.DataBatch(data=(data,))
    mod.forward(db, is_train=False)
    embedding = mod.get_outputs()[0].asnumpy()
    return embedding

def extract_prt_features_img(model_id,output_layer,model,nn,st):
    if output_layer<1:
        img = image.load_img(st, target_size=(nn,nn))
        x = image.img_to_array(img)
        x = np.expand_dims(x, axis=0)
        x = eval('proc'+str(model_id)+'(x)')
        features = model.predict(x)
    else:
        features = get_features(model,st)
    return features


def extract_prt_features(model_id,output_layer,model,img_size,model_name,dirpath,st=''):
    
    img_names = dirfiles(dirpath,'*.png')
    print(output_layer)
    n = len(img_names)
    if output_layer < 1:
        kch = 3
    else:
        kch = 1
    print(kch)
    for i in range(n):
        img_path = img_names[i]
        print(model_name+'_'+st +': '+ num2fixstr(i,4)+'/'+num2fixstr(n,4)+ ': reading '+img_path+'...')
        features = extract_prt_features_img(model_id,output_layer,model,img_size,dirpath+img_path)
        if i==0:
            m = features.shape[kch]
            data = np.zeros((n,m))
            print(data.shape)
            print('size of extracted features:')
            print(features.shape)
        # data[i]    = features[:][0][0] 
        x = np.reshape(features, m, order='F')
        # print(x.shape)    
        data[i]    = x
        # data[i]    = features
    return data









"""
This module includes all the functionalities for geometric analysis in
X-Ray images.

"""

__author__ = 'Christian Pieringer'
__created__ = '2019.08.22'

import numpy as np


def rotation_matrix_2d(theta):
    """
    The 2D rotation matrix R given by a rotation angle theta in radians.

    :param theta: rotation angle in radians (float)
    :returns _r_mat: the rotation matrix (ndarray)
    """

    _r_mat = np.array(
        [
            [np.cos(theta), -np.sin(theta)],
            [np.sin(theta), np.cos(theta)]
        ]
    )

    return _r_mat


def rotation_matrix_3d(wx, wy, wz):
    """
    The method returns the 3D rotation matrix given by a rotation angle theta in radians.

    :param wx: angular rotation in radians for x-axis
    :param wy: angular rotation in radians for y-axis
    :param wz: angular rotation in radians for z-axis
    :returns _r_mat: the rotation matrix for 3D spaces (ndarray)
    """

    _r_mat = np.array([
        [np.cos(wy) * np.cos(wz), -np.cos(wy) * np.sin(wz), np.sin(wy)],
        [np.sin(wx) * np.sin(wy) * np.cos(wz) + np.cos(wx) * np.sin(wz),
         -np.sin(wx) * np.sin(wy) * np.sin(wz) + np.cos(wx) * np.cos(wz),
         -np.sin(wx) * np.cos(wy)],
        [-np.cos(wx) * np.sin(wy) * np.cos(wz) + np.sin(wx) * np.sin(wz),
         np.cos(wx) * np.sin(wy) * np.sin(wz) + np.sin(wx) * np.cos(wz),
         np.cos(wx) * np.cos(wy)]
    ])

    return _r_mat


def get_matrix_p(f):
    """
    Build the 3x4 perspective projection matrix depending on focal distance f.

    Args:
        f: the focal distance (float)

    Returns: the perspective projection matrix p (ndarray)

    """
    p = np.array([
        [f, 0, 0, 0],
        [0, f, 0, 0],
        [0, 0, 1, 0]
    ])
    return p


def hyperproj(m3d, pa, hp):
    """
    Build the 3D->2D transformation using an hyperbolic model (Mery D. and Filbert, D. 2002).

    Mery, D.; Filbert, D. (2002): Automated Flaw Detection in Aluminum Castings Based on The Tracking of Potential
        Defects in a Radioscopic Image Sequence. IEEE Transactions on Robotics and Automation, 18(6):890-901.

    Args:
        m3d:
        pa:
        hp:

    Returns:

    """
    alpha = pa[0]
    u0 = pa[1]
    v0 = pa[2]
    f = pa[3]
    a = pa[4]
    b = pa[5]
    kx = pa[6]
    ky = pa[7]

    mat_a = np.array([
        [kx * np.cos(alpha), ky * np.sin(alpha), u0],
        [-kx * np.sin(alpha), ky * np.cos(alpha), v0],
        [0, 0, 1]]
    )

    mat_b = np.array(
        [
            [f, 0, 0, 0],
            [0, f, 0, 0],
            [0, 0, 1, 0]
        ]
    )

    mat_p = np.dot(mat_b, hp)

    m = np.dot(mat_p, m3d)
    m = m / m[-1]

    xp = m.item(0)
    yp = m.item(1)

    # TODO: check this equation
    c = 1 / np.sqrt((1 - ((xp / a) ** 2)) - ((yp / b) ** 2))

    mpp = np.array([c * xp, c * yp, 1])

    w = np.dot(mat_a, mpp[:, np.newaxis])

    return w




import numpy as np
import matplotlib.pyplot as plt
from sklearn.metrics import confusion_matrix, accuracy_score
from sklearn.model_selection import cross_val_score
from pyxvis.features.selection import fse_model, fsel
from pyxvis.learning.classifiers import define_classifier, train_test_classifier, clf_model


def hold_out(bcl,X,d,Xt,dt):
    ds,clf  = train_test_classifier(bcl,X,d,Xt)
    acc     = accuracy_score(dt,ds)     
    return ds,acc,clf

def cross_validation(bcl,X,d,folds):
    clf    = define_classifier(bcl)
    scores = cross_val_score(clf, X, d, cv=folds)
    acc    = np.mean(scores)
    return acc

def leave_one_out(bcl,X,d):
    classes   = np.sort(np.unique(d))
    dd        = [d[d == di] for di in classes]
    nc        = len(classes)
    nd        = np.zeros((nc,))
    for i in range(nc):
        nd[i] = dd[i].shape[0]
    folds     = int(np.max(nd))
    acc       = cross_validation(bcl,X,d,folds)
    return acc

def best_features_classifier(ss_fs,ff,ss_cl,X,d,Xt,dt,stit,folds=5):
    ncl      = len(ss_cl)           # number of classifiers
    nfs      = len(ss_fs)           # number of feature selectors
    nfk      = len(ff)              # number of feature selectors
    accbest  = 0                    # best achieved accuracy
    acctbest = 0                    # best achieved accuracy
    yacc     = np.zeros((1000,))
    yacct    = np.zeros((1000,))
    nt       = np.zeros((1000,))
    clfs     = [None] * 1000
    fses     = [None] * 1000
    t        = 1

    for k in range(nfk):
        nfik = ff[k]
        for i in range(nfs): 
            fs  = ss_fs[i]
            print('analyzing feature felector: sfs+'+fs+' with '+str(nfik)+' features...')
            (fsname,fspar) = fse_model(fs)
            ss             = fsel([fsname,fspar],X,d,nfik,cv = folds, show = 1)
            Xf             = X[:,ss]
            Xft            = Xt[:,ss]
            for j in range(ncl):
                cl = ss_cl[j]
                # print('analyzing classifier '+cl+'...')
                (clname,clpar) = clf_model(cl)
                acc            = cross_validation([clname,clpar],Xf,d,folds)
                _,acct,_       = hold_out([clname,clpar],Xf,d,Xft,dt)
                stn = f'{nfik:2d}' 
                stt = 'feat='+stn+', fsel='+fs+', clf='+cl
                stb = ''
                sti = '(iter = '+str(t)+')'
                if acc>accbest:
                    accbest = acc
                    stb = sti + ' *** new train max found ***'
                    yacc[t] = acc

                    yacct[t] = acct
                    nt[t] = nfik
                    clfs[t] = cl
                    fses[t] = fs
                    if acct > acctbest:
                        stb = sti + ' *** new train/test max found ***'
                        tbest    = t
                        acctbest = acct
                        accbest0 = acc
                        fsnbest  = fsname
                        fspbest  = fspar
                        fsbest   = fs
                        clbest   = cl
                        clnbest  = clname
                        clpbest  = clpar
                        nbest    = nfik
                        ssbest   = ss

                    t        = t + 1
                print(f'{stt:38s} > acc = {acc:.4f}/{acct:.4f}  <best acc = {accbest:.4f}/{acctbest:.4f}> (Train/Test) '+stb)

    print('---------------------------------------------------------------------------')
    print('       Best iteration: '+str(tbest)+' (maximum of testing accuracy)')
    print('     Feature Selector: ' +fsbest+' with '+str(nbest)+' features')
    print('                     : ('+fsnbest+', '+fspbest+')')
    print('           Classifier: ' +clbest)
    print('                     : ('+clnbest+', '+clpbest+') CrossVal with '+str(folds)+' folds')
    print(f'         Training-Acc: {accbest0:.4f}  ')
    print(f'          Testing-Acc: {acctbest:.4f}  ')
    y1 = yacc[0:t]
    y2 = yacct[0:t]
    x = range(t)
    
    # fig = plt.figure()
    fig = plt.figure(figsize=(t+2,5))
    ax = fig.add_subplot(111)
    plt.plot(x[1:t], y1[1:t], marker='o', markerfacecolor='blue', markersize=6, label = 'Training') 
    plt.plot(x[1:t], y2[1:t], marker='o', markerfacecolor='red', markersize=6, label = 'Testing') 
    plt.xlabel('iterations') 
    plt.ylabel('accuracy') 
    plt.title(stit) 
    yf = 0.405
    dy = 0.05
    dd = 0.02
    fts = 10
    ax.text(dd, yf, 'features:', fontsize=fts)
    ax.text(dd, yf-dy, 'f-selector:', fontsize=fts)
    ax.text(dd, yf-dy*2, 'classifier:', fontsize=fts)
    ax.text(dd, yf-dy*3, 'acc-train:', fontsize=fts)
    ax.text(dd, yf-dy*4, 'acc-test:', fontsize=fts)
    for i in range(1,t):
        col = 'black'
        if i==tbest:
            col = 'green'
        di = i+dd
        ax.text(di, yf, str(int(nt[i])), fontsize=fts,color=col)
        ax.text(di, yf-dy, fses[i], fontsize=fts,color=col)
        ax.text(di, yf-dy*2, clfs[i], fontsize=fts,color=col)
        accs = f'{y1[i]:.4f}'    
        accts = f'{y2[i]:.4f}'    
        ax.text(di, yf-dy*3, accs,  fontsize=fts,color=col)
        ax.text(di, yf-dy*4, accts, fontsize=fts,color=col)
    ax.text(tbest-0.5, yf+0.2, 'best test accuracy', fontsize=fts,color='green')
    ax.annotate(' ',xy=(tbest, acctbest-0.05), xytext=(tbest, yf+0.2),
            arrowprops=dict(facecolor='green', shrink=0.05))
    ax.plot(tbest,acctbest, 'o')
    plt.xticks(range(t+1))    
    plt.grid(True)
    plt.legend()
    plt.xlim(0, t)
    plt.ylim(0, 1.1)
    plt.show() 
    return clbest, ssbest

def precision_recall(dt,ds):
    # dt - true, ds - predicted
    #         0   1
    #     0  TN  FP
    #     1  FN  TP
    # pr = TP/(TP+FP)
    # re = TP/(TP+FN)

    C = confusion_matrix(dt,ds)

    pr = C[1,1]/(C[1,1]+C[0,1])
    re = C[1,1]/(C[1,1]+C[1,0])
    return pr,re




"""
This module implements classes and methods to use the GDXray dataset included in the book.
"""

__author__ = 'Christian Pieringer'
__created__ = '2019.05.08'

import os.path as _path
from os import listdir
from os.path import isdir

import scipy.io as sio
from cv2 import imread, IMREAD_GRAYSCALE
from numpy import genfromtxt, loadtxt, array


# Constants
GDXRAY_PATH = _path.join(_path.expanduser('~'), 'GDXray')


class DatasetBase:
    """
    This class is an abstraction of the images sets class included in this module.
    """

    def __init__(self, group_name, group_prefix, dataset_path=None):
        """
        Init class instance.

        Args:
            dataset_path: string to denote the path of the dataset.

        """

        if not dataset_path:
            root_path = GDXRAY_PATH
        else:
            root_path = dataset_path

        self.root_path = root_path
        self.group_name = group_name
        self.group_prefix = group_prefix
        self.dataset_path = _path.join(self.root_path, self.group_name)

    def __repr__(self):
        class_info = "Class Info:\n"
        class_info += f"Group: {self.group_name}\n"
        class_info += f"Prefix: {self.group_prefix}\n"
        class_info += f"Path: {self.dataset_path}\n"
        class_info += f"Root_path: {self.root_path}\n"
        return class_info

    def describe(self):
        """
        Aggregate dataset information, such as, series names, size, number of images, etc.

        Returns: a python dictionary that aggregates the information.
        """
        # Create a dictionary for count items in each series
        table = {'series': 0, 'images': 0, 'size': 0}

        # Read series within a group
        series = [s for s in listdir(self.dataset_path) if isdir(_path.join(self.dataset_path, s))]

        num_images = 0
        size_bytes = 0

        for s in series:
            images = [i for i in listdir(_path.join(self.dataset_path, s)) if i.endswith('.png')]
            num_images += len(images)

            for f in images:
                size_bytes += _path.getsize(_path.join(self.dataset_path, s, f))

        table['series'] = len(series)
        table['images'] = num_images
        table['size'] = size_bytes / (1024.0 ** 2)

        return table

    def get_dir(self, series):
        """
        Returns the path of an image series.

        Args:
            series (int): the series number

        Returns:
            (str) the series path. It raise an exception if the path does not exist.

        """
        set_path = '{}{:04d}'.format(self.group_prefix, series)
        set_path = _path.join(self.dataset_path, set_path)

        # Check if the directory exist. If not, raise a ValueError exception.
        if not _path.exists(set_path):
            raise ValueError('The series does not exist. Please check the parameters or the root path.')

        return set_path

    def load_gt(self, series, k):
        """
        Load ground truth

        Args:
            series: series number
            k: image number

        Returns:

        """
        _filename = _path.join(self.dataset_path, '{:s}{:04d}'.format(self.group_prefix, series))
        _filename = _path.join(_filename, '{:s}{:04d}_{:04d}.txt'.format(self.group_prefix, series, k))
        ground_truth = genfromtxt('my_file.csv', delimiter=',')

        return ground_truth

    def load_image(self, series, k):
        """
        Loads an image from the database.

        Args:
            series (str): defines the set in the images set.
            k (int): defines the image index.

        Returns:
            (numpy array) loaded image
        """
        _filename = _path.join(self.dataset_path, '{0:s}{1:04d}'.format(self.group_prefix, series))
        _filename = _path.join(_filename, '{0:s}{1:04d}_{2:04d}.png'.format(self.group_prefix, series, k))

        try:
            assert _path.exists(_filename)
        except AssertionError:
            raise Exception('File not found')

        img = imread(_filename, IMREAD_GRAYSCALE)

        return img

    def load_data(self, series, data_type=None):
        """
        Loads an image from GDXray database.

        Args:
            series (str): defines the set in the images set.
            data_type (bool): defining the image index.

        Raises:
            AssertionError: file not found

        Returns:
                _data (numpy array): the output image
        """
        _filename = _path.join(self.dataset_path, '{0}{1:04d}'.format(self.group_prefix, series))
        _filename = _path.join(_filename, '{0}.mat'.format(data_type))

        try:
            assert _path.exists(_filename)
        except AssertionError:
            raise Exception('File not found: {}'.format(_filename))

        # Load the LUT from mat file.
        data = sio.loadmat(_filename)

        _data = None

        if not data_type:
            _data = None
        elif data_type == 'DualEnergyLUT':
            _data = data['LUT']
        elif data_type == 'points':
            _data = data
        else:
            _data = data

        return _data


class Baggages(DatasetBase):

    def __init__(self, group_name='Baggages', group_prefix='B', *args, **kwargs):
        super().__init__(group_name, group_prefix, *args, **kwargs)
        # self.group_name = 'Baggages'
        # self.group_prefix = self.group_names[self.group_name]
        # self.dataset_path = _path.join(self.dataset_path, self.group_name)

    def load_image(self, series, k):
        _filename = _path.join(self.dataset_path, '{:s}{:04d}'.format(self.group_prefix, series))
        _filename = _path.join(_filename, '{:s}{:04d}_{:04d}.png'.format(self.group_prefix, series, k))

        try:
            assert _path.exists(_filename)
        except AssertionError:
            raise Exception('File not found: {}'.format(_filename))

        img = imread(_filename, IMREAD_GRAYSCALE)

        return img

    def load_data(self, series, data_type=None):
        _filename = _path.join(self.dataset_path, 'B{:04d}'.format(series))
        _filename = _path.join(_filename, '{}.mat'.format(data_type))

        try:
            assert _path.exists(_filename)
        except AssertionError:
            raise Exception('File not found')

        # Load the LUT from mat file.
        data = sio.loadmat(_filename)

        _data = None

        if not data_type:
            _data = None
        elif data_type == 'DualEnergyLUT':
            _data = data['LUT']
        else:
            _data = data

        return _data


class Settings(DatasetBase):
    """
    This class implements an instance of the Settings image set.
    """

    def __init__(self, group_name='Settings', group_prefix='S', *args, **kwargs):
        super().__init__(group_name, group_prefix, *args, **kwargs)
        # self.dataset_path = _path.join(self.dataset_path, self.group_name)
        #self.group_prefix = 'S'


class Welds(DatasetBase):
    """
    This class implements an instance of the Settings image set.
    """

    def __init__(self, group_name='Welds', group_prefix='W', *args, **kwargs):
        super().__init__(group_name, group_prefix, *args, **kwargs)
        #self.dataset_path = _path.join(self.dataset_path, 'Welds')
        # self.group_prefix = 'W'


class Nature(DatasetBase):
    """
    This class implements an instance of the Settings image set.
    """

    def __init__(self, group_name='Nature', group_prefix='N', *args, **kwargs):
        super().__init__(group_name, group_prefix, *args, **kwargs)
        # self.group_prefix = 'N'
        # self.group_name = self.group_names[self.group_prefix]
        #self.dataset_path = _path.join(self.dataset_path, self.group_name)


class Castings(DatasetBase):
    """
    This class implements an instance of the Settings image set.
    """

    def __init__(self, group_name='Castings', group_prefix='C', *args, **kwargs):
        super().__init__(group_name, group_prefix, *args, **kwargs)
        #self.dataset_path = _path.join(self.dataset_path, 'Castings')
        # self.group_prefix = 'C'

    def load_data(self, series, data_type=None):
        # In this case the method loads the file defined as argument.
        _filename = _path.join(self.dataset_path, '{0}{1:04d}'.format(self.group_prefix, series))
        _filename = _path.join(_filename, '{0}'.format(data_type))

        data = loadtxt(_filename)

        return data


def load_image_set(set_name, *args, **kwargs):
    image_sets = {
        'baggages': Baggages,
        'settings': Settings,
        'nature': Nature,
        'welds': Welds,
        'castings': Castings
    }

    return image_sets[set_name.lower()](*args, **kwargs)


class ImageSet(DatasetBase):
    """
    This class create a generic image set container.
    """

    def __init__(self, *args, **kwargs):
        """

        :param collection_dir:
        """
        super().__init__(*args, **kwargs)
        # print(self.dataset_path)

    def load_image(self, serie_name, k):
        """

        :param serie_name: serie name (str)
        :param k: image index (int)
        :return: ndarray with the image
        """
        _filename = _path.join(self.dataset_path, serie_name)
        _filename = _path.join(_filename, '{:s}_{:04d}.png'.format(serie_name, k))

        try:
            assert _path.exists(_filename)
        except AssertionError:
            raise Exception('File not found')

        img = imread(_filename, IMREAD_GRAYSCALE)

        return img


def xgdx_stats(root_dir=None):
    """
    Compute a complete set of statistics on the dataset, such as, the total number of groups, the total number of
    series per group, and the size of each group in MB.

    Args:
        root_dir:

    Raises:
        ValueError: dataset directory not found

    Returns:
        table (dict): a summary of dataset statistics.
    """

    if not root_dir:
        root_dir = GDXRAT_PATH

    # Check if the directory exist. If not, raise a ValueError exception.
    if not _path.exists(root_dir):
        raise ValueError('Directory not found. Please check the root path.')

    # Create a dictionary for count items in each series
    table = {
        g:
            {'series': 0,
             'images': 0,
             'size': 0
             }
        for g in listdir(root_dir)
    }

    for g in table:
        series_dir = _path.join(root_dir, g)
        series = [d for d in listdir(series_dir) if
                  _path.isdir(_path.join(series_dir, d))]

        num_images = 0
        size_bytes = 0

        for s in series:
            images = [i for i in listdir(_path.join(root_dir, g, s)) if
                      i.endswith('png')]
            num_images += len(images)

            for f in images:
                size_bytes += _path.getsize(_path.join(root_dir, g, s, f))

        table[g]['series'] = len(series)
        table[g]['images'] = num_images
        table[g]['size'] = size_bytes / (1024.0 ** 2)

    return table


def load_gt(img_set, series, k, gt_filename='ground_truth.txt'):
    """
    Load ground truth from a series

    Args:
        img_set (object): XGDX image set instance
        series: series number
        k: image number
        gt_filename:

    Returns:

    """
    _filename = _path.join(img_set.dataset_path, '{:s}{:04d}'.format(img_set.group_prefix, series))
    _filename = _path.join(_filename, gt_filename)

    # Check if the series contains ground truth
    if not _path.isfile(_filename):
        raise IOError('Ground truth file not found')

    _ground_truth = genfromtxt(_filename)
    _ground_truth = _ground_truth[_ground_truth[:, 0] == k][:, 1:]

    return _ground_truth


if __name__ == '__main__':

    from pyxvis.io import gdxraydb

    # Testing the class instantiation
    set_names = ['Castings', 'Welds', 'Nature', 'Settings', 'Baggages']

    for set_name in set_names:
        image_set = load_image_set(set_name)
        print(image_set)
        print(image_set.describe())

    # Testing loading ground truth
    image_set = gdxraydb.Baggages()
    img = image_set.load_image(2, 4)



import os
import numpy as np
import cv2
from keras.layers                    import Dense,GlobalAveragePooling2D
from keras.models                    import Model
from keras.callbacks                 import ModelCheckpoint, EarlyStopping
from keras.preprocessing.image       import ImageDataGenerator
from keras.optimizers                import Adam, SGD
from keras.applications              import MobileNet, MobileNetV2, VGG16, VGG19
from keras.applications.resnet50     import ResNet50
from keras.applications.xception     import Xception
#from keras.applications.densenet     import DenseNet121
#from keras.applications.nasnet       import NASNetLarge, NASNetMobile
from keras.applications.inception_v3 import InceptionV3
from pyxvis.io.misc import dirfiles
from pyxvis.io.plots import plot_confusion

from sklearn.metrics import confusion_matrix, accuracy_score


def tfl_model(base_model):
    if base_model == 1: # MobileNet (87 layers)
        model_name = 'MobileNet'
        bmodel     = MobileNet(weights='imagenet',include_top=False) 
    elif base_model == 2: # InceptionV3 (311 layers)
        model_name = 'InceptionV3'
        bmodel     = InceptionV3(weights='imagenet',include_top=False) 
    elif base_model == 3: # VGG16 (16 layers)
        model_name = 'VGG16'
        bmodel     = VGG16(weights='imagenet',include_top=False) 
    elif base_model == 4: # VGG119 (22 layers)
        model_name = 'VGG19'
        bmodel     = VGG19(weights='imagenet',include_top=False) 
    elif base_model == 5: # ResNet50 (175 layers)
        model_name = 'ResNet50'
        bmodel     = ResNet50(weights='imagenet',include_top=False) 
    elif base_model == 6: # Xception (132 layers)
        model_name = 'Xception'
        bmodel     = Xception(weights='imagenet',include_top=False) 
    elif base_model == 7: # MobileNetV2 (155 layers)
        model_name = 'MobileNetV2'
        bmodel     = MobileNetV2(weights='imagenet',include_top=False) 
    #elif base_model == 8: # DenseNet121 (427 layers)
    #    model_name = 'DenseNet121'
    #    bmodel     = DenseNet121(weights='imagenet',include_top=False) 
    #elif base_model == 9: # NASNetMobile (769 layers)
    #    model_name = 'NASNetMobile'
    #    bmodel     = NASNetMobile(weights='imagenet',include_top=False) 
    #elif base_model == 10: # NASNetLarge (1039 layers)
    #    model_name = 'NASNetLarge'
    #    bmodel     = NASNetLarge(weights='imagenet',include_top=False) 
    else:
        print('error: base model '+str(base_model)+' not defined.')
    bmodel.summary()
    print('\nBase model '+ model_name + ' loaded with '+str(len(bmodel.layers)) + ' layers.')
    return bmodel


def tfl_define_model(bmodel,dense_layers,NumClasses):
    x = bmodel.output
    x = GlobalAveragePooling2D()(x)
    n = len(dense_layers)
    for i in range(n):
        layers_i = dense_layers[i]
        if layers_i > 0:
            x = Dense(layers_i,activation='relu')(x)

    preds = Dense(NumClasses,activation='softmax')(x) #final layer with softmax activation
    model = Model(inputs=bmodel.input,outputs=preds)
    return model


def generate_training_set(val_split, augmentation, batch_size, path_dataset, img_size):
    # augmentation is how aggressive will be the data augmentation/transformation, eg. 0.05
    if val_split > 0:
        image_generator = ImageDataGenerator(rescale=1. / 255,
                                                    rotation_range   = augmentation,
                                                    shear_range      = augmentation,
                                                    zoom_range       = augmentation,
                                                    cval             = augmentation,
                                                    horizontal_flip  = True,
                                                    vertical_flip    = True,
                                                    validation_split = val_split)

        print('\ndefining training subset from '+path_dataset+'/train...')
        train_set   = image_generator.flow_from_directory(batch_size  = batch_size,
                                                    directory   = path_dataset+'/train',
                                                    shuffle     = True,
                                                    target_size = img_size, 
                                                    subset      = "training",
                                                    class_mode  = 'categorical')

        print('\ndefining validation subset from '+path_dataset+'/train...')
        val_set     = image_generator.flow_from_directory(batch_size  = batch_size,
                                                    directory   = path_dataset+'/train',
                                                    shuffle     = True,
                                                    target_size = img_size, 
                                                    subset      = "validation",
                                                    class_mode  = 'categorical')
    else:
        train_generator = ImageDataGenerator(rescale=1. / 255,
                                                    rotation_range  = augmentation,
                                                    shear_range     = augmentation,
                                                    zoom_range      = augmentation,
                                                    cval            = augmentation,
                                                    horizontal_flip = True,
                                                    vertical_flip   = True)

        val_generator = ImageDataGenerator(rescale=1. / 255)

        print('\ndefining training subset from '+path_dataset+'/train...')
        train_set = train_generator.flow_from_directory(path_dataset+'/train',
                                                    target_size = img_size,
                                                    batch_size  = batch_size,
                                                    class_mode  = 'categorical')
        print('\ndefining validation subset from '+path_dataset+'/val...')
        val_set = val_generator.flow_from_directory(path_dataset+'/val',
                                                    target_size = img_size,
                                                    batch_size  = batch_size,
                                                    class_mode  = 'categorical')

    return train_set, val_set

def defineCallBacks(model_file):
    callbacks = [
        EarlyStopping(
            monitor        = 'val_acc', 
            patience       = 8,
            mode           = 'max',
            verbose        = 1),
        ModelCheckpoint(model_file,
            monitor        = 'val_acc', 
            save_best_only = True, 
            mode           = 'max',
            verbose        = 0)
    ]
    return callbacks

def deleteWeights(best_model,last_model):
    if os.path.exists(best_model):
        os.remove(best_model)
    if os.path.exists(last_model):
        os.remove(last_model)

def tfl_train(bmodel,model,opti_method,nb_layers,train_set,train_steps,val_set,val_steps,nb_epochs,path_dataset,nb_classes,img_size):

    best_model   = 'tfl_best_model.h5'
    last_model   = 'tfl_last_model.h5'
    top_model    = 'tfl_top_model.h5'

    nb = len(bmodel.layers)
    nm = len(model.layers)

    if nb_layers < 0:
        nb_layers = nb + nb_layers

    print('\nBase model has '+str(nb)+' layers, the first '+str(nb_layers)+' ones are frozen.')

    if nb_layers > nb:
        print('Error: frezzed layers exceeds the total number of layers.')

    accmax = 0

    for j in range(3):
    
        if nb_epochs[j] > 0:

            if j==0:
                st = 'After Training-1 (new layers)'
                print('\nTraining on only new layers (top '+str(nm-nb)+' layers): '+str(nb_epochs[j])+' epochs')
                for layer in bmodel.layers:
                    layer.trainable = False
            elif j==1:
                st = 'After Training-2 (not frozen + new layers)'
                print('\nTraining on not frozen layers and ew layers (top '+str(nm-nb_layers)+' layers): '+str(nb_epochs[j])+' epochs')
                for layer in model.layers[:nb_layers]:
                    layer.trainable = False
                for layer in model.layers[nb_layers:]:
                    layer.trainable = True
            elif j==2:
                st = 'After Training-3 (all layers)'
                print('\nTraining on all layers ('+str(nm)+' layers): '+str(nb_epochs[j])+' epochs')
                for layer in model.layers:
                    layer.trainable = True

            if opti_method == 1: # Adam
                model.compile(optimizer='Adam',loss='categorical_crossentropy',metrics=['accuracy'])
            else: # SGD
                model.compile(optimizer=SGD(lr=0.0001, momentum=0.9), loss='categorical_crossentropy',metrics=['accuracy'])

            callbacks    = defineCallBacks(best_model)
            deleteWeights(best_model,last_model)

            model.fit_generator(
                train_set,
                steps_per_epoch  = train_steps,
                validation_data  = val_set, 
                validation_steps = val_steps,
                epochs           = nb_epochs[j],
                callbacks        = callbacks
            )

            model.save_weights(last_model)
            print('loading best model from '+best_model+' ...')
            model.load_weights(best_model)
            print('\n'+st+':')
            C,acc = tfl_testing_accuracy(model,path_dataset,nb_classes,img_size)
            if acc>accmax:
                print('*** top model achieved '+st+' ***')
                jmax     = j
                accmax   = acc
                Cmax     = C
                topmodel = model
                print('saving top model: '+top_model+' ...')
                model.save_weights(top_model)
            plot_confusion(C,acc,st,0,nb_classes)
    print('Top model achieved after Training-'+str(jmax+1))
    print('Confusion matrix:')
    print(Cmax)
    accst = f'Acc = {acc:.4f}'    
    print(accst)
    return topmodel,Cmax,accmax


def tfl_classify_image(model,imgpath,img_size,show):
    
    img = cv2.imread(imgpath)
    img = cv2.resize(img,(img_size[0],img_size[1]))
    img = np.reshape(img,[1,img_size[0],img_size[1],3])/255
    x = model.predict(img)
    cl = np.argmax(x[0])
    if show == 1:
        print(imgpath+' > class '+str(cl) +': ' + '%6.4f' % x[0][cl])
    return cl


def tfl_testing_accuracy(model,path_dataset,nb_classes,img_size):
    print('\nEvaluation on testing dataset ('+str(nb_classes)+' classes):')
    ygt = np.ones([0,1]) # ground truth
    y   = np.ones([0,1]) # predictions
    nt  = 0
    for i in range(nb_classes):
        path_i = path_dataset+'/test/class_'+str(i)+'/'
        img_test_i = dirfiles(path_i,'*.jpg')
        n_i    = len(img_test_i)
        nt     = nt + n_i
        ygt_i  = i*np.ones([n_i,1])
        ygt    = np.concatenate((ygt, ygt_i), axis=0)
        y_i    = np.zeros([n_i,1])
        print('   > class '+str(i)+'/'+str(nb_classes-1)+': evaluating model on '+str(n_i)+' images of '+path_i+' ...')
        for j in range(n_i):
            img_path = img_test_i[j]
            y_i[j]   = tfl_classify_image(model,path_i+img_path,img_size,0)
        y  = np.concatenate((y, y_i), axis=0)

    print(str(nt)+' test images evaluated.')
 
    C   = confusion_matrix(ygt, y) 
    acc = accuracy_score(ygt,y)

    return C, acc






import numpy as np
from scipy.io import loadmat
from keras.utils.np_utils import to_categorical
from sklearn.model_selection import train_test_split


def load_features(prefix,full=0,categorical=0):
    X  = np.load(prefix+'_Xtrain.npy')            # training samples
    Xt = np.load(prefix+'_Xtest.npy')             # testing samples
    d  = np.ravel(np.load(prefix+'_dtrain.npy'))  # training labels
    dt = np.ravel(np.load(prefix+'_dtest.npy'))   # testing labels
    if full==0:
        print('Training data: '+str(X.shape[0]) +' samples with '+str(X.shape[1]) +' features')
        print(' Testing data: '+str(Xt.shape[0])+' samples with '+str(Xt.shape[1])+' features')
        print('      Classes: '+str(int(np.min(d)))+'...'+str(int(np.max(d))))
        if categorical==1:
            dmin = np.min(d)
            Ytrain = np.transpose(to_categorical(d-dmin))
            Ytest  = np.transpose(to_categorical(dt-dmin))
            Xtrain = np.transpose(X)
            Xtest  = np.transpose(Xt)
            return Xtrain,Ytrain,Xtest,Ytest
        else:
            return X,d,Xt,dt
    else:
        X = np.concatenate((X,Xt),axis=0)
        d = np.concatenate((d,dt),axis=0)
        print('Available data: '+str(X.shape[0]) +' samples with '+str(X.shape[1]) +' features')
        print('       Classes: '+str(int(np.min(d)))+'...'+str(int(np.max(d))))
        if categorical==1:
            dmin = np.min(d)
            Y = np.transpose(to_categorical(d-dmin))
            X = np.transpose(X)
            return X,Y
        else:
            return X,d

def save_features(X,d,Xt,dt,st):
    print('Training data (X,d)  : '+str(X.shape[0]) +' samples with '+str(X.shape[1]) +' features')
    print(' Testing data (Xt,dt): '+str(Xt.shape[0])+' samples with '+str(Xt.shape[1])+' features')
    print('      Classes: '+str(int(np.min(d)))+'...'+str(int(np.max(d))))
    print('saving training data (X,d) and testing data (Xt,dt)...')
    print('... training features in ' +st+'_Xtrain.npy')
    np.save(st+'_Xtrain',X)
    print('... training labels   in ' +st+'_dtrain.npy')
    np.save(st+'_dtrain',d)
    print('... testing  features in ' +st+'_Xest.npy')
    np.save(st+'_Xtest',Xt)
    print('... testing  labels   in ' +st+'_dest.npy')
    np.save(st+'_dtest',dt)



def load_cnn_patches(st_file):
    # patches are of 32 x 32 pixels (uint8)
    # classes are two: 0 and 1
    # X_train and X_test are uint8 4D matrices of size N_train x 1 x 32 x 32 and N_test x 1 x 32 x 32 respectively
    # Y_train and Y_test are uint8 column vectors with 0 and 1 of size N_train x 1 and N_test x 1 respectively
    print('loading training/testing data from '+ st_file +' ...')
    if st_file[-4:]=='.mat':
        data              = loadmat(st_file)
        X_train  = data['X_train']
        Y_train  = data['Y_train']
        X_test   = data['X_test']
        Y_test   = data['Y_test']
    else:
        X_train  = np.load(st_file+'_Xtrain.npy')
        Y_train  = np.load(st_file+'_Ytrain.npy')
        X_test   = np.load(st_file+'_Xtest.npy')
        Y_test   = np.load(st_file+'_Ytest.npy')
    
    Y_train  = to_categorical(Y_train)
    print('X train size: {}'.format(X_train.shape))
    print('y train size: {}'.format(Y_train.shape))

    Y_test   = to_categorical(Y_test)
    print('X test  size: {}'.format(X_test.shape))
    print('y test  size: {}'.format(Y_test.shape))

    classes  = [0, 1]

    return X_train, Y_train, X_test, Y_test, classes

def load_cnn_test_patches(st_file):
    print('loading testing data from '+ st_file +' ...')
    if st_file[-4:]=='.mat':
        data              = loadmat(st_file)
        X_test   = data['X_test']
        Y_test   = data['Y_test']
    else:
        X_test   = np.load(st_file+'_Xtest.npy')
        Y_test   = np.load(st_file+'_Ytest.npy')
    Y_test   = to_categorical(Y_test)
    print('X test  size: {}'.format(X_test.shape))
    print('y test  size: {}'.format(Y_test.shape))

    classes  = [0, 1]

    return X_test, Y_test, classes

def init_data(X,k):
    d = k*np.ones([X.shape[0],],dtype=int)
    return X,d

def append_data(X,d,Xa,ja):
    (Xa,da) = init_data(Xa,ja)
    X = np.concatenate((X,Xa),axis=0)
    d = np.concatenate((d,da),axis=0)
    return X,d

